# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def _l2_norm_fwd_1pass_kernel(
    X,  # pointer to the input
    Y,  # pointer to the output
    stride_x_row,  # how much to increase the pointer when moving by 1 row
    N,  # number of columns in X
    eps,  # epsilon to avoid division by zero
    BLOCK_N: tl.constexpr,
):
    # Map the program id to the row of X and Y it should compute.
    row = tl.program_id(0)
    X += row * stride_x_row
    Y += row * stride_x_row
    # Compute mean and variance
    cols = tl.arange(0, BLOCK_N)
    x = tl.load(X + cols, mask=cols < N, other=0.0).to(tl.float32)
    xbar = tl.where(cols < N, x, 0.0)
    var = tl.sum(xbar * xbar, axis=0) 
    rstd = 1 / tl.sqrt(var + eps)
    # Normalize and apply linear transformation
    mask = cols < N
    y = x * rstd
    # Write output
    tl.store(Y + cols, y, mask=mask)

@triton.jit
def _l2_norm_bwd_kernel(
    X,  # pointer to the input
    DY,  # pointer to the output gradient
    DX,  # pointer to the input gradient
    stride_x_row,  # how much to increase the pointer when moving by 1 row
    N,  # number of columns in X
    eps,  # epsilon to avoid division by zero
    BLOCK_N: tl.constexpr,
):
    # Map the program id to the row of X and Y it should compute.
    row = tl.program_id(0)
    X += row * stride_x_row
    DX += row * stride_x_row
    DY += row * stride_x_row

    cols = tl.arange(0, BLOCK_N)
    x = tl.load(X + cols, mask=cols < N, other=0.0).to(tl.float32)
    x = tl.where(cols < N, x, 0.0)
    var = tl.sum(x * x) 
    rstd = 1 / tl.sqrt(var + eps)
    # Normalize and apply linear transformation
    mask = cols < N
    dy = tl.load(DY + cols, mask=cols < N, other=0.0).to(tl.float32)
    dy = tl.where(cols < N, dy, 0.0)
    dx = dy * rstd - tl.sum(dy * x) * (1 / (var+eps)) * rstd * x
    tl.store(DX + cols, dx, mask=mask)

def _l2_norm_fwd(
    x, eps=1e-6
):
    x_shape_og = x.shape
    x = x.reshape(-1, x.shape[-1])
    if x.stride(-1) != 1:
        x = x.contiguous()
    assert x.stride(-1) == 1 
    # allocate output
    y = torch.empty_like(x)
    assert y.stride(-1) == 1
    N = x.shape[-1]
    M = x.shape[0]
    # Less than 64KB per feature: enqueue fused kernel
    MAX_FUSED_SIZE = 65536 // x.element_size()
    BLOCK_N = min(MAX_FUSED_SIZE, triton.next_power_of_2(N))
    if N > BLOCK_N:
        raise RuntimeError(
            "This layer norm doesn't support feature dim >= 64KB.")
    # heuristics for number of warps
    with torch.cuda.device(x.device.index):
        _l2_norm_fwd_1pass_kernel[(M,)](
            x,
            y,
            x.stride(0),
            N,
            eps,
            BLOCK_N,
        )
    return y.reshape(x_shape_og)

def _l2_norm_bwd(
    x, dy, eps=1e-5,
):
    x_shape_og = x.shape
    x = x.reshape(-1, dy.shape[-1])
    dy = dy.reshape(-1, dy.shape[-1])
    if dy.stride(-1) != 1:
        dy = dy.contiguous()
    assert dy.shape == x.shape
    # allocate output
    dx = torch.empty_like(x)
    N = x.shape[-1]
    M = x.shape[0]
    assert x.stride(-1) == 1
    assert dy.stride(-1) == 1
    # Less than 64KB per feature: enqueue fused kernel
    MAX_FUSED_SIZE = 65536 // x.element_size()
    BLOCK_N = min(MAX_FUSED_SIZE, triton.next_power_of_2(N))
    if N > BLOCK_N:
        raise RuntimeError(
            "This layer norm doesn't support feature dim >= 64KB.")
    # heuristics for number of warps
    with torch.cuda.device(x.device.index):
        _l2_norm_bwd_kernel[(M,)](
            x,
            dy,
            dx,
            x.stride(0),
            N,
            eps,
            BLOCK_N,
        )
    return dx.reshape(x_shape_og)




##################################################################################################################################################


import torch

def test_l2_norm_triton():
    # Test parameters
    batch_size, dim = 8, 128  # Define dimensions for test tensor
    eps = 1e-6

    # Initialize input tensor
    x = torch.randn((batch_size, dim), dtype=torch.float32, device="cuda", requires_grad=True)
    
    # Dictionary to store test results
    test_results = {}

    # Forward pass test
    y = _l2_norm_fwd(x, eps=eps)
    test_results["test_case_1"] = y

    # Backward pass test
    dy = torch.ones_like(y, device="cuda")
    dx_analytical = _l2_norm_bwd(x, dy, eps=eps)
    test_results["test_case_2"] = dx_analytical

    return test_results

result_gold = test_l2_norm_triton()



import sys
import os
import json

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.token_attn_mistral import token_att_fwd2
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('token_attn_mistral', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 16):  # Adjust the range as needed for your testing
            batch_size = 2 ** i
            seq_len = 128
            head = 8  # Example head size
            dim = 64  # Example dimension size
            sliding_window = 64

            prob = torch.rand((head, seq_len), dtype=torch.float16)
            v = torch.rand((head, seq_len, dim), dtype=torch.float16)
            out = torch.zeros((batch_size, head, dim), dtype=torch.float16)
            Req_to_tokens = torch.randint(0, seq_len, (batch_size, seq_len), dtype=torch.int32)
            B_req_idx = torch.randint(0, batch_size, (batch_size,), dtype=torch.int32)
            B_Start_Loc = torch.zeros((batch_size,), dtype=torch.int32)
            B_Seqlen = torch.full((batch_size,), seq_len, dtype=torch.int32)
            B_Att_Start_Loc = torch.zeros((batch_size,), dtype=torch.int32)
            B_Att_Seqlen = torch.full((batch_size,), seq_len, dtype=torch.int32)

            self.input_tensors.append((prob, v, out, Req_to_tokens, B_req_idx, B_Start_Loc, B_Seqlen, B_Att_Start_Loc, B_Att_Seqlen, sliding_window))

    def to_cuda(self, input_tensor):
        # return tuple(tensor.cuda() for tensor in input_tensor)
        prob, v, out, Req_to_tokens, B_req_idx, B_Start_Loc, B_Seqlen, B_Att_Start_Loc, B_Att_Seqlen, sliding_window = input_tensor
        return (prob.cuda(), v.cuda(), out.cuda(), Req_to_tokens.cuda(), B_req_idx.cuda(), B_Start_Loc.cuda(), B_Seqlen.cuda(), B_Att_Start_Loc.cuda(), B_Att_Seqlen.cuda(), sliding_window)

    def call_op(self, input_tensor):
        prob, v, out, Req_to_tokens, B_req_idx, B_Start_Loc, B_Seqlen, B_Att_Start_Loc, B_Att_Seqlen, sliding_window = input_tensor
        return token_att_fwd2(prob, v, out, Req_to_tokens, B_req_idx, B_Start_Loc, B_Seqlen, B_Att_Start_Loc, B_Att_Seqlen, sliding_window)

    def get_gbps(self, input_tensor, runtime):
        prob, v, out, _, _, _, _, _, _, _ = input_tensor
        total_bytes = (prob.numel() + v.numel() + out.numel()) * prob.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        prob, v, _, _, _, _, _, _, _, _ = input_tensor
        FLOPS = 2 * prob.numel() * v.size(-1)  # Assuming each element in prob contributes to a FLOP
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



import sys
import os
import json

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.fifth_order_sph_harmonics import FifthOrderSphericalHarmonic
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('fifth_order_sph_harmonics', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 20):
            size = 2 ** i
            input_tensor = torch.rand(size, 3, dtype=torch.float16)  # Assuming 3D coordinates
            self.input_tensors.append(input_tensor)

    def to_cuda(self, input_tensor):
        return input_tensor.cuda()

    def call_op(self, input_tensor):
        output_tensor = FifthOrderSphericalHarmonic.apply(input_tensor)
        return output_tensor

    def get_gbps(self, input_tensor, runtime):
        x = input_tensor
        total_bytes = x.numel() * x.element_size() + 11 * x.size(0) * x.element_size()  # Input and output
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        x = input_tensor
        # Assuming each element in the output involves a certain number of FLOPs
        FLOPS = 11 * x.size(0) * 100  # Example: 100 FLOPs per output element
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS
    
    def run_benchmark(self):
        results = []
        for input_tensor_ in self.input_tensors:
            input_tensor = self.to_cuda(input_tensor_)
            # print(input_tensor)
            op = lambda : self.call_op(input_tensor)
            ms = self.get_runtime(op)
            gbps = self.get_gbps(input_tensor, ms)
            tflops = self.get_tflops(input_tensor, ms)
            result = {
                "input_size": [input_tensor.shape],
                "ms": ms,
                "GB/s": gbps,
                "TFLOPS": tflops
            }
            print(result)
            results.append(result)
            input_tensor = None
        folder_path = "/home/lishangzhan/triton/bench_performance/results"
        file_name = self.op_name + ".json"
        file_path = os.path.join(folder_path, file_name)
        with open(file_path, 'w', encoding='utf8') as f:
            json.dump(results, f, indent=4)

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.destindex_copy_kv1 import destindex_copy_kv
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('destindex_copy_kv1', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 18):  # Example sizes, adjust as needed
            seq_len = 2 ** i
            head_num = 64  # Example head number
            head_dim = 64  # Example head dimension
            K = torch.rand(seq_len, head_num, head_dim, dtype=torch.float16)
            DestLoc = torch.randint(0, seq_len, (seq_len,), dtype=torch.int32)
            Out = torch.zeros_like(K)
            self.input_tensors.append((K, DestLoc, Out))

    def to_cuda(self, input_tensor):
        K, DestLoc, Out = input_tensor
        return (K.cuda(), DestLoc.cuda(), Out.cuda())

    def call_op(self, input_tensor):
        K, DestLoc, Out = input_tensor
        destindex_copy_kv(K, DestLoc, Out)
        return Out

    def get_gbps(self, input_tensor, runtime):
        K, DestLoc, Out = input_tensor
        total_bytes = (K.numel() + DestLoc.numel() + Out.numel()) * K.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        K, DestLoc, Out = input_tensor
        FLOPS = 2 * K.numel()  # Assuming each element involves a load and a store
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Copyright(C) [2025] Advanced Micro Devices, Inc. All rights reserved.
import triton  
import triton.language as tl  


@triton.jit  
def moe_gemm_kernel(  
    A,  
    B,  
    Out,  
    A_scale,  
    B_scale,  
    stride_am,  
    stride_ak,  
    stride_be,  
    stride_bn,  
    stride_bk,  
    stride_cm,  
    stride_cn,  
    stride_bse,  
    stride_bsn,  
    top_k: tl.constexpr,  
    topk_weights_ptr,  
    sorted_token_ids_ptr,  
    expert_ids_ptr,  
    EM: tl.constexpr,  
    N: tl.constexpr,  
    K: tl.constexpr,  
    EVEN_K: tl.constexpr,  
    MUL_ROUTED_WEIGHT: tl.constexpr,  
    use_fp8_w8a8: tl.constexpr,  
    use_int8_w8a16: tl.constexpr,  
    use_int8_w8a8: tl.constexpr,  
    BLOCK_SIZE_M: tl.constexpr,  
    BLOCK_SIZE_N: tl.constexpr,  
    BLOCK_SIZE_K: tl.constexpr,  
    GROUP_SIZE_M: tl.constexpr,  
):  
    """  
    Implements the fused computation for a Mixture of Experts (MOE) using  
    token and expert matrices.  
  
    Key Parameters:  
    - A: The input tensor representing tokens with shape (*, K), where '*' can  
        be any shape representing batches and K is the feature dimension of  
        each token.  
    - B: The stacked MOE weight tensor with shape (E, N, K), where E is  
        the number of experts, K is the input feature dimension, and N is  
        the output feature dimension.  
    - C: The output cache tensor with shape (M, topk, N), where M is the  
        total number of tokens post padding, topk is the number of times  
        each token is repeated, and N is the output feature dimension.  
    - sorted_token_ids: A tensor containing the sorted indices of tokens,  
        repeated topk times and arranged by the expert index they are  
        assigned to.  
    - expert_ids: A tensor containing the indices of the expert for each  
        block. It determines which expert matrix from B should be used for  
        each block in A.  
    This kernel performs the multiplication of a token by its corresponding  
    expert matrix as determined by `expert_ids`. The sorting of  
    `sorted_token_ids` by expert index and padding ensures divisibility by  
    BLOCK_SIZE_M, which is necessary to maintain consistency in block matrix  
    multiplication across different blocks processed by the same expert.  
    """  
    pid = tl.program_id(axis=0)  
    num_pid_m = tl.cdiv(EM, BLOCK_SIZE_M)  
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)  
    num_pid_in_group = GROUP_SIZE_M * num_pid_n  
    group_id = pid // num_pid_in_group  
    first_pid_m = group_id * GROUP_SIZE_M  
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)  
    pid_m = first_pid_m + ((pid % num_pid_in_group) % group_size_m)  
    pid_n = (pid % num_pid_in_group) // group_size_m  
  
    offs_token_id = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)  
    offs_token = tl.load(sorted_token_ids_ptr + offs_token_id)  
  
    # Here we assume that valid tokens are in the range [0, M).  
    token_mask = (offs_token >= 0) & (offs_token < EM)  
  
    off_experts = tl.load(expert_ids_ptr + pid_m)  
    offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N).to(tl.int64)) % N  
    offs_k = tl.arange(0, BLOCK_SIZE_K)  
    a_ptrs = A + (offs_token[:, None] // top_k * stride_am + offs_k[None, :] * stride_ak)  
    b_ptrs = B + off_experts * stride_be + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)  
  
    if use_int8_w8a16:  
        b_scale_ptrs = B_scale + off_experts * stride_bse + offs_bn[None, :] * stride_bsn  
        b_scale = tl.load(b_scale_ptrs)  
  
    if use_fp8_w8a8 or use_int8_w8a8:  
        a_scale = tl.load(A_scale)  
        b_scale = tl.load(B_scale + off_experts)  
  
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)  
  
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):  
        # Masking ensures we don't load from invalid tokens or indices  
        if EVEN_K:  
            a = tl.load(a_ptrs, mask=(token_mask[:, None]), other=0.0)  
            b = tl.load(b_ptrs)  
        else:  
            a = tl.load(a_ptrs, mask=(token_mask[:, None] & (offs_k[None, :] < K - k * BLOCK_SIZE_K)), other=0.0)  
            b = tl.load(b_ptrs, mask=(offs_k[:, None] < K - k * BLOCK_SIZE_K), other=0.0)  
  
        if use_int8_w8a16:  
            accumulator = tl.dot(a, b.to(a.dtype), acc=accumulator)  
        elif use_fp8_w8a8 or use_int8_w8a8:  
            accumulator += tl.dot(a, b)  
        else:  
            accumulator = tl.dot(a, b, acc=accumulator)  
  
        a_ptrs += BLOCK_SIZE_K * stride_ak  
        b_ptrs += BLOCK_SIZE_K * stride_bk  
  
    if MUL_ROUTED_WEIGHT:  
        moe_weight = tl.load(topk_weights_ptr + offs_token, mask=token_mask, other=0)  
        accumulator = accumulator * moe_weight[:, None]  
  
    if use_int8_w8a16:  
        accumulator = (accumulator * b_scale).to(Out.dtype.element_ty)  
    elif use_fp8_w8a8 or use_int8_w8a8:  
        accumulator = (accumulator * a_scale * b_scale).to(Out.dtype.element_ty)  
    else:  
        accumulator = accumulator.to(Out.dtype.element_ty)  
  
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)  
    out_ptrs = Out + stride_cm * offs_token[:, None] + stride_cn * offs_cn[None, :]  
    c_mask = token_mask[:, None] & (offs_cn[None, :] < N)  
    tl.store(out_ptrs, accumulator.to(Out.dtype.element_ty), mask=c_mask)


##################################################################################################################################################

  
import triton # Required for triton.testing utilities and launching kernel  
import torch  
import pytest  
from typing import Any, Dict, Optional  
import os  
import json  
import functools  
import argparse  
import sys  
import triton.language as tl # Required for tl.constexpr in quantize_input and other places  


from geak_eval.perf.ROCm.performance_utils_pytest import PytestBenchmarker, do_bench_config, save_all_benchmark_results
from typing import Dict
######################################## HELPERS for Eval ######################################## 
import numpy as np
import random
import torch 

result_gold = {}
CONFIG = {
  "llama3": {
    "8B": {
      "num_attention_heads": 32,
      "num_key_value_heads": 8,
      "hidden_size": 4096,
      "intermediate_size": 14336,
      "vocab_size": 128256
    },
    "70B": {
      "num_attention_heads": 64,
      "num_key_value_heads": 8,
      "hidden_size": 8192,
      "intermediate_size": 28672,
      "vocab_size": 128256
    },
    "405B": {
      "num_attention_heads": 128,
      "num_key_value_heads": 8,
      "hidden_size": 16384,
      "intermediate_size": 53248,
      "vocab_size": 128256
    }
  },
  "mistral": {
    "7B": {
      "hidden_size": 4096,
      "intermediate_size": 14336,
      "num_attention_heads": 32,
      "num_key_value_heads": 8,
      "vocab_size": 32000
    },
    "22B": {
      "hidden_size": 6144,
      "intermediate_size": 16384,
      "num_attention_heads": 48,
      "num_key_value_heads": 8,
      "vocab_size": 32000
    }

  }
}

  
def get_model_configs(config_path='model_configs.json', model_families=["llama3"], model="all"):  
    """  
    Load model names from the configuration file.  
  
    Args:  
        config_path (str): User-provided path to the configuration JSON file.  
        model_families (list): List of model family names to retrieve.  
  
    Returns:  
        dict: A dictionary of available models and their configurations for the specified families.  
    """  
    configs = CONFIG.copy()
  
    # Extract models and their configurations for the specified families  
    filtered_configs = {}  
  
    for family in model_families:  
        if family in configs:  
            # Check if model filtering is required  
            if model == "all":  
                # Include all models in the family  
                for model_size, model_configs in configs[family].items():  
                    filtered_configs[f"{family}-{model_size}"] = model_configs  
            else:  
                # Parse the model string (e.g., llama3_8B or llama3-8B)  
                delimiter = "_" if "_" in model else "-"  
                model_parts = model.split(delimiter)  
  
                # Check if the family and size match  
                if len(model_parts) == 2 and model_parts[0] == family:  
                    model_size = model_parts[1]  
                    if model_size in configs[family]:  
                        filtered_configs[f"{family}-{model_size}"] = configs[family][model_size]  
  
    if not filtered_configs:  
        print(f"Warning: No models selected for families: {model_families} with filter: '{model}'")  
  
    return filtered_configs  
  
  
def get_available_models(config_file='model_configs.json', model_families=["llama3"]):  
    """  
    Load model names from the configuration file.  
  
    Args:  
        config_file (str): Path to the configuration JSON file.  
        model_families (list): List of model family names to retrieve.  
  
    Returns:  
        list: A list of available models for the specified families.  
    """  
    configs = CONFIG.copy()
  
    models = [f"{family}-{model}" for family in model_families if family in configs for model in configs[family]]  
  
    return models  


def set_seed(seed: int = 42) -> None:
    """
    Set the random seed for reproducibility across multiple libraries and configure PyTorch for deterministic behavior.

    Args:
        seed (int): The seed value to set. Default is 42.
    """
    # Set seed for Python's built-in random module
    random.seed(seed)
    # Set seed for NumPy
    np.random.seed(seed)
    # Set seed for PyTorch on CPU
    torch.manual_seed(seed)
    # Set seed for PyTorch on all GPUs (if available)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    # Ensure deterministic behavior in PyTorch
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # Set environment variable for hash-based operations
    os.environ['PYTHONHASHSEED'] = str(seed)

######################################## HELPERS for Eval ######################################## 


M_THRESHOLD_SMALL = 256  
M_THRESHOLD_MEDIUM = 1024  
  
dtype_max = {  
    dtype: (torch.finfo(dtype) if dtype.is_floating_point else torch.iinfo(dtype)).max  
    for dtype in [  
        torch.float8_e5m2fnuz,  
        torch.float8_e4m3fnuz,  
        torch.int8,  
    ]  
}  
  
supported_fp8 = [torch.float8_e4m3fnuz, torch.float8_e5m2fnuz]  
  
  
class MetaData():  
    use_fp8_w8a8 = False  
    use_int8_w8a16 = False  
    use_int8_w8a8 = False  
  
    def __init__(self, top_k, topk_weights, topk_ids, sorted_token_ids, expert_ids, num_tokens_post_padded, config):  
        self.top_k = top_k  
        self.topk_weights = topk_weights  
        self.topk_ids = topk_ids  
        self.sorted_token_ids = sorted_token_ids  
        self.expert_ids = expert_ids  
        self.num_tokens_post_padded = num_tokens_post_padded  
        self.config = config  
  
    def set_use_fp8_w8a8(self, a_descale, b_descale, fp8_type):  
        self.use_fp8_w8a8 = True  
        self.a_descale = a_descale  
        self.b_descale = b_descale  
        self.fp8_type = fp8_type  
  
    def set_use_int8_w8a16(self, b_descale):  
        self.use_int8_w8a16 = True  
        self.b_descale = b_descale  
        self.a_descale = None  
  
    def set_use_int8_w8a8(self, a_descale, b_descale):  
        self.use_int8_w8a8 = True  
        self.a_descale = a_descale  
        self.b_descale = b_descale  
  
    def check_args(self, a, b, o):  
        assert a.shape[-1] == b.shape[-1] and b.shape[1] == o.shape[-1]  
  
        assert not (self.use_fp8_w8a8 and self.use_int8_w8a16 and self.use_int8_w8a8)  
        if self.use_fp8_w8a8:  
            assert self.fp8_type in supported_fp8, f"fp8 type {self.fp8_type} not supported"  
  
  
def _moe_align_block_size(topk_ids: torch.Tensor, num_experts: int, top_k: int, block_size: int,  
                          sorted_token_ids: torch.Tensor, expert_ids: torch.Tensor,  
                          num_tokens_post_pad: torch.Tensor) -> None:  
    M, top_k = topk_ids.shape  
  
    expert_to_tokens = [[] for _ in range(num_experts)]  
    # For each token, for each selected expert, we append (token_id, expert)  
    for token_id in range(M):  
        for j in range(top_k):  
            e_id = topk_ids[token_id, j].item()  
            expert_to_tokens[e_id].append(token_id * top_k + j)  
  
    # Reorder tokens block by block, padding if needed  
    reordered_token_ids = []  
    reordered_expert_ids = []  
  
    for e_id in range(num_experts):  
        tokens_for_expert = expert_to_tokens[e_id]  
        num_tokens = len(tokens_for_expert)  
  
        n_blocks = ((num_tokens + block_size - 1) // block_size)  
        # If not a multiple of block_size, pad up to the next multiple  
        padded_size = n_blocks * block_size  
  
        # Reorder all actual tokens for expert e_id  
        reordered_token_ids.extend(tokens_for_expert)  
        # reordered_expert_ids.extend([e_id]*num_tokens)  
        reordered_expert_ids.extend([e_id] * n_blocks)  
  
        # Pad with dummy token_id = -1 (or any sentinel), if needed  
        if padded_size > num_tokens:  
            pad_count = padded_size - num_tokens  
            reordered_token_ids.extend([-1] * pad_count)  
  
    token_length = len(reordered_token_ids)  
    expert_length = len(reordered_expert_ids)  
  
    sorted_token_ids[:token_length] = torch.tensor(reordered_token_ids, dtype=sorted_token_ids.dtype,  
                                                   device=sorted_token_ids.device)  
    expert_ids[:expert_length] = torch.tensor(reordered_expert_ids, dtype=expert_ids.dtype, device=expert_ids.device)  
  
    # Fill remainder with -1 if these arrays are bigger than total_length  
    if token_length < sorted_token_ids.numel():  
        sorted_token_ids[token_length:] = -1  
    if expert_length < expert_ids.numel():  
        expert_ids[expert_length:] = -1  
  
    num_tokens_post_pad.fill_(token_length)  
  
  
def moe_align_block_size(topk_ids: torch.Tensor, block_size: int,  
                         num_experts: int) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:  
    """  
    Aligns the token distribution across experts to be compatible with block size for matrix multiplication.  
  
    Parameters:  
    - topk_ids: A tensor of shape [total_tokens, top_k] representing the top-k expert indices for each token.  
    - block_size: The block size used in block matrix multiplication.  
    - num_experts: The total number of experts.  
  
    Returns:  
    - sorted_token_ids: A tensor containing the sorted token indices according to their allocated expert.  
    - expert_ids: A tensor indicating the assigned expert index for each block.  
    - num_tokens_post_padded: The total number of tokens after padding, ensuring divisibility by block_size.  
  
    This function pads the number of tokens that each expert needs to process so that it is divisible by block_size.  
    Padding ensures that during block matrix multiplication, the dimensions align correctly.  
  
    Example:  
    Given topk_ids = [[2, 3, 4], [1, 2, 4], [1, 3, 4], [1, 2, 3]], block_size = 4, and num_experts = 4:  
    - We initially have 12 tokens (after repeating 'top_k' times) and 4 experts, with each expert needing to process 3 tokens.  
    - As block_size is 4, we pad 1 token for each expert.  
    - First, flatten topk_ids to [2, 3, 4, 1, 2, 4, 1, 3, 4, 1, 2, 3].  
    - Then append padding tokens [12, 12, 12, 12] for each block.  
    - After sorting by expert index, we obtain token_ids [3, 6, 9, 12, 0, 4, 10, 12, 1, 7, 11, 12, 2, 5, 8, 12].  
        Tokens 12 are non-existent (padding) and are ignored in the subsequent matrix multiplication.  
    - The padding ensures that the total number of tokens is now divisible by block_size for proper block matrix operations.  
    """  
    top_k = topk_ids.shape[1]  
    sorted_ids = torch.empty((topk_ids.numel() + num_experts * (block_size - 1), ), dtype=torch.int32,  
                             device=topk_ids.device)  
    expert_ids = torch.empty((topk_ids.numel() + num_experts, ), dtype=torch.int32, device=topk_ids.device)  
    sorted_ids.fill_(topk_ids.numel())  
    num_tokens_post_pad = torch.empty((1), dtype=torch.int32, device=topk_ids.device)  
    _moe_align_block_size(topk_ids, num_experts, top_k, block_size, sorted_ids, expert_ids, num_tokens_post_pad)  
  
    return sorted_ids, expert_ids, num_tokens_post_pad  
  
  
def get_config_dtype_str(dtype: torch.dtype, use_int8_w8a16: Optional[bool] = False,  
                         use_int8_w8a8: Optional[bool] = False, use_fp8_w8a8: Optional[bool] = False):  
    if use_fp8_w8a8:  
        return "fp8_w8a8"  
    elif use_int8_w8a16:  
        return "int8_w8a16"  
    elif use_int8_w8a8:  
        return "int8_w8a8"  
    elif dtype == torch.float:  
        # avoiding cases where kernel fails when float32 MoE  
        # use fp16/bfloat16 configs  
        return "float32"  
    return None  
  
  
def get_config_file_name(dtype: Optional[str]) -> str:  
    device_name = torch.cuda.get_device_name(0).replace(" ", "_")  
    dtype_selector = "" if not dtype else f",dtype={dtype}"  
    return f"device_name={device_name}{dtype_selector}.json"  
  
  
@functools.lru_cache  
def get_moe_configs(dtype: Optional[str]) -> Optional[Dict[int, Any]]:  
    """  
    Return optimized configurations for the fused MoE kernel.  
  
    The return value will be a dictionary that maps an irregular grid of  
    batch sizes to configurations of the fused_moe kernel. To evaluate the  
    kernel on a given batch size bs, the closest batch size in the grid should  
    be picked and the associated configuration chosen to invoke the kernel.  
    """  
    # First look up if an optimized configuration is available in the configs  
    # directory  
    json_file_name = get_config_file_name(dtype)  
  
    config_file_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "configs", json_file_name)  
    if os.path.exists(config_file_path):  
        with open(config_file_path) as f:  
            # If a configuration has been found, return it  
            return {key: val for key, val in json.load(f).items()}  
  
    # If no optimized configuration is available, we will use the default  
    # configuration  
    return None  
  
  
def get_default_config(  
    M: int,  
    E: int,  
    is_marlin: bool,  
) -> Dict[str, int]:  
    config = {'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}  
    # A heuristic: fused marlin works faster with this config for small M  
    if M <= E or (is_marlin and M <= 32):  
        config = {'BLOCK_SIZE_M': 16, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 64, 'GROUP_SIZE_M': 1}  
    return config  
  
  
def try_get_optimal_moe_config(  
    E: int,  
    dtype: Optional[str],  
    M: int,  
    is_marlin: bool = False,  
):  
    configs = get_moe_configs(dtype)  
  
    if configs:  
        if configs: # This inner 'if configs:' is redundant  
            if M < M_THRESHOLD_SMALL:  
                config = configs["small_M"]  
            elif M < M_THRESHOLD_MEDIUM:  
                config = configs["medium_M"]  
            else:  
                config = configs["large_M"]  
    else:  
        # Else use the default config  
        config = get_default_config(M, E, is_marlin)  
  
    return config  
  
  
def moe_gemm(a: torch.Tensor, b: torch.Tensor, c: torch.Tensor, metadata: MetaData) -> None:  
    # TODO shard M dim  
    metadata.check_args(a, b, c)  
  
    num_tokens_post_padded, topk_weights, sorted_token_ids, expert_ids, config = metadata.num_tokens_post_padded, metadata.topk_weights, metadata.sorted_token_ids, metadata.expert_ids, metadata.config  
  
    use_fp8_w8a8, use_int8_w8a16, use_int8_w8a8 = metadata.use_fp8_w8a8, metadata.use_int8_w8a16, metadata.use_int8_w8a8  
    a_descale, b_descale = None, None  
    stride_bse = None  
    stride_bsn = None  
    if use_fp8_w8a8 or use_int8_w8a16 or use_int8_w8a8:  
        a_descale, b_descale = metadata.a_descale, metadata.b_descale  
        if use_int8_w8a16:  
            stride_bse = b_descale.stride(0)  
            stride_bsn = b_descale.stride(1)  
  
    top_k = metadata.top_k  
  
    EM = num_tokens_post_padded.item()  
    _, N, K = b.shape  
    grid = lambda META: (triton.cdiv(EM, META['BLOCK_SIZE_M']) * triton.cdiv(N, META['BLOCK_SIZE_N']), )  
  
    EVEN_K = K % config["BLOCK_SIZE_K"] == 0  
      
    # This is where the kernel defined above is called  
    # We need to ensure moe_gemm_kernel is in scope, which it is if defined globally or imported.  
    # For this re-structuring, it's assumed the kernel from the <triton-kernel-code> block is accessible.  
    # If running this directly, you might need to ensure the kernel definition is executed first.  
    moe_gemm_kernel[grid](a, b, c, a_descale,  
                          b_descale, a.stride(0), a.stride(1), b.stride(0), b.stride(1), b.stride(2), c.stride(1),  
                          c.stride(2), stride_bse, stride_bsn, top_k, topk_weights, sorted_token_ids, expert_ids, EM, N,  
                          K, EVEN_K, MUL_ROUTED_WEIGHT=topk_weights is not None, use_fp8_w8a8=use_fp8_w8a8,  
                          use_int8_w8a16=use_int8_w8a16, use_int8_w8a8=use_int8_w8a8, **config)  
    return c  
  
  
def quantize_tensor(tensor: torch.Tensor, dtype, dim=()) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:  
    quantize_dim = [i for i in range(tensor.dim()) if i not in dim]  
    max_vals = tensor.abs().amax(dim=quantize_dim, keepdim=True)  
    max_repr_val = dtype_max[dtype]  
    # Avoid division by zero  
    max_vals[max_vals == 0] = 1e-8  
  
    # Compute scale factors for each channel  
    scale: torch.Tensor = max_repr_val / max_vals.to(torch.float32)  
  
    # Quantize the tensor  
    tensor = tensor * scale  
    if dtype == torch.int8:  
        tensor = tensor.round_()  
    tensor.clamp_(-max_repr_val, max_repr_val)  
    tensor_quantized = tensor.to(dtype)  
  
    scale = scale.squeeze(dim=quantize_dim)  
  
    return tensor_quantized, scale, 1 / scale  
  
  
def quantize_input(a, b, use_fp8_w8a8: tl.constexpr, use_int8_w8a16: tl.constexpr, use_int8_w8a8: tl.constexpr,  
                   metatdata: MetaData, fp8_type=None):  
    assert not (use_fp8_w8a8 and use_int8_w8a16 and use_int8_w8a8)  
    assert not (use_fp8_w8a8 and fp8_type is None)  
  
    if use_fp8_w8a8:  
        a_quantized, _, a_descale = quantize_tensor(a, dtype=fp8_type)  
        b_quantized, _, b_descale = quantize_tensor(b, dim=(0, ), dtype=fp8_type)  
        metatdata.set_use_fp8_w8a8(a_descale, b_descale, fp8_type)  
        return a_quantized, b_quantized  
  
    if use_int8_w8a8:  
        a_quantized, _, a_descale = quantize_tensor(a, dtype=torch.int8)  
        b_quantized, _, b_descale = quantize_tensor(b, dim=(0, ), dtype=torch.int8)  
        metatdata.set_use_int8_w8a8(a_descale, b_descale)  
        return a_quantized, b_quantized  
  
    if use_int8_w8a16:  
        b_quantized, _, b_descale = quantize_tensor(b, dim=(0, 1), dtype=torch.int8)  
        metatdata.set_use_int8_w8a16(b_descale)  
        return a, b_quantized  
  
  
def input_helper(M: int, N: int, K: int, top_k: int, E: int, routed_weight: bool, use_fp8_w8a8: bool,  
                 use_int8_w8a16: bool, use_int8_w8a8: bool, fp8_type, dtype):  

    set_seed()
    a = torch.randn((M, K), dtype=dtype, device='cuda')  
    b = torch.randn((E, N, K), dtype=dtype, device='cuda')  
    c = torch.zeros((M, top_k, N), dtype=dtype, device='cuda')  
  
    values = torch.randn(M, E, device='cuda')  
  
    softmax_vals = torch.softmax(values, dim=1)  
    topk_weights, topk_ids = torch.topk(softmax_vals, k=top_k, dim=1)  
  
    config_dtype = get_config_dtype_str(use_fp8_w8a8=use_fp8_w8a8, use_int8_w8a16=use_int8_w8a16,  
                                        use_int8_w8a8=use_int8_w8a8, dtype=dtype)  
    get_config_func = functools.partial(  
        try_get_optimal_moe_config,  
        E,  
        config_dtype,  
    )  
    config = get_config_func(M)  
    sorted_token_ids, expert_ids, num_tokens_post_padded = moe_align_block_size(topk_ids, config['BLOCK_SIZE_M'], E)  
  
    metadata = MetaData(top_k, topk_weights if routed_weight else None, topk_ids, sorted_token_ids, expert_ids,  
                        num_tokens_post_padded, config)  
  
    if use_fp8_w8a8 or use_int8_w8a16 or use_int8_w8a8:  
        a, b = quantize_input(a, b, use_fp8_w8a8, use_int8_w8a16, use_int8_w8a8, metadata, fp8_type)  
  
    return a, b, c, metadata  
  
  
@pytest.mark.parametrize("M, N, K, top_k, E", [  
    (64, 14336, 4096, 2, 8),  
    (16, 14336, 1, 2, 4),  
    (1, 14336, 128, 2, 4),  
    (3, 14336, 128, 2, 4),  
    (16, 14336, 128, 1, 4),  
    (16, 14336, 128, 1, 1),  
    (64, 7186, 128, 2, 8),  
    (64, 3584, 128, 2, 8),  
    (64, 1792, 128, 2, 8),  
    (64, 64, 128, 2, 8),  
])  
@pytest.mark.parametrize('routed_weight', [True, False])  
def test_correctness(M: int, N: int, K: int, top_k: int, E: int, routed_weight: bool, request, dtype=torch.float16):  
    
    a, b, c, metadata = input_helper(M, N, K, top_k, E, routed_weight=routed_weight, use_fp8_w8a8=False,  
                                     use_int8_w8a16=False, use_int8_w8a8=False, fp8_type=None, dtype=dtype)  
  
    tri_out = moe_gemm(a, b, c, metadata)  

    
    ################### save tri_out in result_gold ###################
    test_case_name = request.node.name
    sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
    result_gold[sanitized_key_name] = tri_out.clone().detach().cpu()
    ###################################################################
    
    topk_ids = metadata.topk_ids  
    topk_weights = metadata.topk_weights  
    ref_out = torch.empty_like(c)  
    # Repeat a -> (M, top_k, K)  
    a_expanded = a.unsqueeze(1).repeat(1, top_k, 1)  
    # (M, top_k, N, K)  
    b_indexed = b[topk_ids]  
    ref_out = torch.einsum("mek,menk->men", a_expanded, b_indexed)  
    if routed_weight:  
        ref_out *= topk_weights.unsqueeze(-1)  

    result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
    # Validate correctness  
    torch.testing.assert_close(tri_out, ref_out, atol=1e-2, rtol=1e-2)  
  
  
@pytest.mark.parametrize("M, N, K, top_k, E", [  
    (64, 14336, 4096, 2, 8),  
    (16, 14336, 1, 2, 4),  
    (1, 14336, 128, 2, 4),  
    (16, 14336, 128, 1, 4),  
    (16, 14336, 128, 1, 1),  
    (64, 7186, 128, 2, 8),  
    (64, 3584, 128, 2, 8),  
    (64, 1792, 128, 2, 8),  
    (64, 64, 128, 2, 8),  
])  
@pytest.mark.parametrize('routed_weight', [True, False])  
@pytest.mark.parametrize('use_fp8_w8a8', [True])  
@pytest.mark.parametrize('fp8_type', [torch.float8_e4m3fnuz, torch.float8_e5m2fnuz])  
def test_correctness_fp8(M: int, N: int, K: int, top_k: int, E: int, routed_weight: bool, use_fp8_w8a8, fp8_type, request, dtype=torch.float16): 

    a, b, c, metadata = input_helper(M, N, K, top_k, E, routed_weight=routed_weight, use_fp8_w8a8=use_fp8_w8a8,  
                                     use_int8_w8a16=False, fp8_type=fp8_type, use_int8_w8a8=False, dtype=dtype)  
  
    tri_out = moe_gemm(a, b, c, metadata)

    result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
    ################### save tri_out in result_gold ###################
    test_case_name = request.node.name
    sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
    result_gold[sanitized_key_name] = tri_out.clone().detach().cpu()
    ###################################################################

    topk_ids = metadata.topk_ids  
    topk_weights = metadata.topk_weights  
    ref_out = torch.empty_like(c)  
    # Repeat a -> (M, top_k, K)  
    a_expanded = a.unsqueeze(1).repeat(1, top_k, 1)  
    # (M, top_k, N, K)  
    b_indexed = b.half()[topk_ids]  
    ref_out = torch.einsum("mek,menk->men", a_expanded.float(), b_indexed.float())  
  
    if routed_weight:  
        ref_out *= topk_weights.unsqueeze(-1)  
  
    ref_out = ref_out * metadata.b_descale[topk_ids].unsqueeze(-1)  
    ref_out = ref_out * metadata.a_descale  
    ref_out = ref_out.to(dtype)  
  
    # Validate correctness  
    torch.testing.assert_close(tri_out, ref_out, atol=1e-2, rtol=1e-2)  
  
  
@pytest.mark.parametrize("M, N, K, top_k, E", [  
    (64, 14336, 4096, 2, 8),  
    (16, 14336, 1, 2, 4),  
    (1, 14336, 128, 2, 4),  
    (16, 14336, 128, 1, 4),  
    (16, 14336, 128, 1, 1),  
    (64, 7186, 128, 2, 8),  
    (64, 3584, 128, 2, 8),  
    (64, 1792, 128, 2, 8),  
    (64, 64, 128, 2, 8),  
])  
@pytest.mark.parametrize('routed_weight', [True, False])  
@pytest.mark.parametrize('use_int8_w8a16', [True])  
def test_correctness_int8_w8a16(M: int, N: int, K: int, top_k: int, E: int, routed_weight: bool, use_int8_w8a16, request,  
                                dtype=torch.float16):  
    a, b, c, metadata = input_helper(M, N, K, top_k, E, routed_weight=routed_weight, use_fp8_w8a8=False,  
                                     use_int8_w8a16=use_int8_w8a16, use_int8_w8a8=False, fp8_type=None, dtype=dtype)  
  
    tri_out = moe_gemm(a, b, c, metadata)  
    result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
    ################### save tri_out in result_gold ###################
    test_case_name = request.node.name
    sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
    result_gold[sanitized_key_name] = tri_out.clone().detach().cpu()
    ###################################################################
    
    topk_ids = metadata.topk_ids  
    topk_weights = metadata.topk_weights  
    ref_out = torch.empty_like(c)  
    # Repeat a -> (M, top_k, K)  
    a_expanded = a.unsqueeze(1).repeat(1, top_k, 1)  
    # (M, top_k, N, K)  
    b_indexed = b[topk_ids]  
    ref_out = torch.einsum("mek,menk->men", a_expanded.float(), b_indexed.float())  
    if routed_weight:  
        ref_out *= topk_weights.unsqueeze(-1)  
  
    ref_out = ref_out * metadata.b_descale[topk_ids, :]  
    ref_out = ref_out.to(dtype)  
  
    # Validate correctness  
    torch.testing.assert_close(tri_out, ref_out, atol=1e-2, rtol=1e-2)  
  
  
@pytest.mark.parametrize("M, N, K, top_k, E", [  
    (64, 14336, 4096, 2, 8),  
    (16, 14336, 1, 2, 4),  
    (1, 14336, 128, 2, 4),  
    (16, 14336, 128, 1, 4),  
    (16, 14336, 128, 1, 1),  
    (64, 7186, 128, 2, 8),  
    (64, 3584, 128, 2, 8),  
    (64, 1792, 128, 2, 8),  
    (64, 64, 128, 2, 8),  
])  
@pytest.mark.parametrize('routed_weight', [True, False])  
@pytest.mark.parametrize('use_int8_w8a8', [True])  
def test_correctness_int8_w8a8(M: int, N: int, K: int, top_k: int, E: int, routed_weight: bool, use_int8_w8a8, request,  
                               dtype=torch.float16):    
    a, b, c, metadata = input_helper(M, N, K, top_k, E, routed_weight=routed_weight, use_fp8_w8a8=False,  
                                     use_int8_w8a16=False, use_int8_w8a8=use_int8_w8a8, fp8_type=None, dtype=dtype)  
  
    tri_out = moe_gemm(a, b, c, metadata)  
    result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
    ################### save tri_out in result_gold ###################
    test_case_name = request.node.name
    sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
    result_gold[sanitized_key_name] = tri_out.clone().detach().cpu()
    ###################################################################
    
    topk_ids = metadata.topk_ids  
    topk_weights = metadata.topk_weights  
    ref_out = torch.empty_like(c)  
    # Repeat a -> (M, top_k, K)  
    a_expanded = a.unsqueeze(1).repeat(1, top_k, 1)  
    # (M, top_k, N, K)  
    b_indexed = b[topk_ids]  
    ref_out = torch.einsum("mek,menk->men", a_expanded.float(), b_indexed.float())  
    if routed_weight:  
        ref_out *= topk_weights.unsqueeze(-1)  
  
    ref_out = ref_out * metadata.b_descale[topk_ids].unsqueeze(-1)  
    ref_out = ref_out * metadata.a_descale  
    ref_out = ref_out.to(dtype)  
  
    
    # Validate correctness  
    torch.testing.assert_close(tri_out, ref_out, atol=1e-2, rtol=1e-2)  

# --- Define TFLOPS and GB/s calculators for MoE GEMM ---
def calculate_moe_gemm_tflops(params: dict, ms: float) -> float:
    M = params['M_orig'] # Original number of tokens before top_k expansion
    N = params['N']
    K = params['K']
    top_k = params['top_k']
    # Each of M tokens interacts with top_k experts.
    # For each such interaction, it's an M_slice * K @ K * N GEMM.
    # Effective operations: M * top_k * (2 * K * N)
    flops = M * top_k * (2 * K * N)
    if params.get('routed_weight', False):
        flops += M * top_k * N # Element-wise multiplication by routing weights
    tflops = flops / (ms / 1000) / 1e12
    return tflops

def calculate_moe_gemm_gbps(params: dict, ms: float) -> float:
    M_orig = params['M_orig']
    N = params['N']
    K = params['K']
    E = params['E']
    top_k = params['top_k']
    dtype_str = params.get('dtype_str', 'fp16') # Default if not specified

    if dtype_str == 'fp32': current_dtype = torch.float32
    elif dtype_str == 'bf16': current_dtype = torch.bfloat16
    else: current_dtype = torch.float16
    element_size = torch.tensor([], dtype=current_dtype).element_size()

    # Memory I/O:
    # Read A: (M_orig, K)
    # Read B (all expert weights): (E, N, K) - worst case, all loaded
    # Read routing info: topk_ids (M_orig, top_k), expert_ids (complex, related to padded M), topk_weights (M_orig, top_k)
    # Write C: (M_orig, top_k, N) -> effectively M_orig * top_k rows of size N are written.

    bytes_a = M_orig * K * element_size
    bytes_b = E * N * K * element_size # All expert weights
    # Output c_for_kernel is (EM_padded, N). EM_padded is roughly M_orig * top_k.
    # Let's use M_orig * top_k for a cleaner estimate of useful data written.
    bytes_c_written = M_orig * top_k * N * element_size

    # Routing info bytes are usually smaller and sometimes omitted for simplicity,
    # but can be significant for small M, N, K.
    # topk_ids: M_orig * top_k * 4 (int32)
    # expert_ids: (num_blocks_for_kernel) * 4 (int32) - harder to estimate precisely without full alignment logic
    # topk_weights: M_orig * top_k * element_size (if routed_weight)
    bytes_routing = M_orig * top_k * 4 # topk_ids
    if params.get('routed_weight', False):
        bytes_routing += M_orig * top_k * element_size

    # For MoE, a common simplification is A + B_active + C
    # B_active = M_orig * top_k * K * N (effectively) but data comes from (E,N,K)
    # Let's use: Read A, Read all B, Write C_useful
    total_bytes = bytes_a + bytes_b + bytes_c_written # Add bytes_routing if significant
    gbps = total_bytes / (ms / 1000) / 1e9
    return gbps

# --- Name for the benchmark output JSON file ---
OP_NAME_FOR_BENCHMARK = "moe_gemm_triton_perf"

# --- Pytest parametrize for performance testing (based on test_correctness) ---
MOE_GEMM_PARAMS_FOR_PERF = [
    # M, N, K, top_k, E
    (64, 14336, 4096, 2, 8),
    (256, 7168, 4096, 2, 8), # Example medium size
    (1024, 14336, 4096, 2, 8), # Example larger M
    (16, 1024, 512, 1, 4),   # Smaller K, N
]
MOE_DTYPES_FOR_PERF = ['fp16', 'bf16'] # Reduced set for faster perf testing
# Quantization modes are complex to integrate here simply, focus on main dtypes first.

@pytest.mark.parametrize("M_orig, N, K, top_k, E", MOE_GEMM_PARAMS_FOR_PERF)
@pytest.mark.parametrize('routed_weight', [True, False])
@pytest.mark.parametrize('dtype_str', MOE_DTYPES_FOR_PERF)
def test_performance(M_orig, N, K, top_k, E, routed_weight, dtype_str, request):
    # Determine torch dtype
    set_seed()
    if dtype_str == 'fp32': current_dtype = torch.float32
    elif dtype_str == 'bf16': current_dtype = torch.bfloat16
    else: current_dtype = torch.float16

    # --- Input Setup using input_helper ---
    # For performance, we are not testing quantization variants like fp8/int8 here for simplicity.
    # That would require more parameters for use_fp8_w8a8, etc.
    a, b, c_for_kernel, metadata = input_helper(
        M=M_orig, N=N, K=K, top_k=top_k, E=E,
        routed_weight=routed_weight,
        use_fp8_w8a8=False, use_int8_w8a16=False, use_int8_w8a8=False,
        fp8_type=None, dtype=current_dtype
    )

    # --- Create op_lambda for benchmarking ---
    op_lambda = lambda: moe_gemm(a, b, c_for_kernel, metadata)

    # --- Benchmarking ---
    bench_config = do_bench_config(warm_up=10, repetition=50) # MoE can be slower
    benchmarker = PytestBenchmarker(op_callable=op_lambda,
                                    op_name=OP_NAME_FOR_BENCHMARK,
                                    config=bench_config)

    current_params_for_logs_and_calc = {
        "M_orig": M_orig, "N": N, "K": K, "top_k": top_k, "E": E,
        "routed_weight": routed_weight, "dtype_str": dtype_str,
        # Include relevant metadata.config if it affects performance
        "BLOCK_SIZE_M": metadata.config['BLOCK_SIZE_M'],
        "BLOCK_SIZE_N": metadata.config['BLOCK_SIZE_N'],
        "BLOCK_SIZE_K": metadata.config['BLOCK_SIZE_K'],
        "GROUP_SIZE_M": metadata.config['GROUP_SIZE_M'],
    }

    benchmarker.run_benchmark(current_params_dict=current_params_for_logs_and_calc,
                              gbps_calculator=calculate_moe_gemm_gbps,
                              tflops_calculator=calculate_moe_gemm_tflops)
    
######################################## HELPERS for Eval ########################################     
# --- Pytest hook to save the dictionary at the end of the session ---  
def test_save_results():  
    """  
    Called after whole test run finished, right before returning the exit status to the system.  
    """
    print('Inside session finish...')
    if "_CALL_SUCCESS_" not in result_gold:
        result_gold['_CALL_SUCCESS_'] = torch.tensor([[0.0]])
    OUTPUT_FILENAME = __file__.replace('.','_') + '.pt'
    print(f"\nSaving all tri_out results to {OUTPUT_FILENAME}...")  
    # Ensure the directory for the output file exists if it's in a subdirectory  
    output_dir = os.path.dirname(OUTPUT_FILENAME)  
    if output_dir and not os.path.exists(output_dir):  
        os.makedirs(output_dir, exist_ok=True)  
    torch.save(result_gold, OUTPUT_FILENAME)       
    print(f"Successfully saved {len(result_gold)} tri_out tensors to {OUTPUT_FILENAME}.")  

def test_save_performance_results():
    """
    Called after the test_performance function finishes.
    This is a separate hook to ensure performance results are saved.
    """
    print('\nPytest session finishing... Saving benchmark results...')

    output_directory = os.path.join(os.path.dirname(__file__), "perf")  # Save in a "perf" subdirectory next to the test file
    os.makedirs(output_directory, exist_ok=True)
    
    save_all_benchmark_results(output_directory)
    print(f"All benchmark results attempted to save to: {output_directory}")
######################################## HELPERS for Eval ######################################## 


def get_configs():  
    configs = [  
        {"M": 64, "N": 256, "K": 128, "E": 8, "top_k": 2},  
        {"M": 64, "N": 1792, "K": 1024, "E": 8, "top_k": 2},  
        {"M": 64, "N": 7168, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 128, "N": 7168, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 1024, "N": 7168, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 4096, "N": 7168, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 64, "N": 14336, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 128, "N": 14336, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 256, "N": 14336, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 512, "N": 14336, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 1024, "N": 14336, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 2048, "N": 14336, "K": 4096, "E": 8, "top_k": 2},  
        {"M": 4096, "N": 14336, "K": 4096, "E": 8, "top_k": 2},  
    ]  
    return configs  
  
  
def model_benchmark_configs(args):  
    config_file = args.model_configs  
    configs = get_model_configs(config_path=config_file, model_families=["mistral"], model=args.model)  
    moe_configs = []  
    M = args.M if args.M else 4096  # check size  
    # M, K, N, E, top_k  
  
    for model_name, config in configs.items():  
        N1 = config["intermediate_size"]  
        K1 = config["hidden_size"]  
  
        N2 = config["hidden_size"]  
        K2 = config["intermediate_size"] // 2  
  
        E = 8  
        top_k = 2  
        # The first moe layer  
        moe_configs.append((model_name, M, N1, K1, E, top_k))  
        # The second moe layer  
        moe_configs.append((model_name, M * top_k, N2, K2, E, 1))  
  
    return moe_configs  
  
  
def run_benchmark(custom, args):  
    routed_weight = args.routed_weight  
    use_int8_w8a16 = args.int8_w8a16  
    use_fp8_w8a8 = args.fp8_w8a8  
    use_int8_w8a8 = args.int8_w8a8  
    dtype = arg_to_torch_dtype[args.dtype]  
    fp8_type = arg_to_torch_dtype[args.fp8_type]  
  
    x_names = ['M', 'N', 'K', 'E', 'top_k']  
    if custom:
        assert args.M and args.N and args.K and args.E and args.top_k, \
            "Please provide M, N, K, E, top_k for custom runs."
        x_vals_list = [(args.M, args.N, args.K, args.E, args.top_k)] 
    else:  
        if args.model:  
            x_vals_list = model_benchmark_configs(args)  
            x_names = ['model', 'M', 'N', 'K', 'E', 'top_k']  
        else:  
            configs = get_configs()  
            x_vals_list = [(cfg['M'], cfg['N'], cfg['K'], cfg['E'], cfg['top_k']) for cfg in configs]  
  
    line_names = ['Time (ms)', 'TFLOPS', 'Bandwidth (GB/s)']  
    line_vals = ['time', 'tflops', 'bandwidth']  
  
    benchmark = triton.testing.Benchmark(  
        x_names=x_names, x_vals=x_vals_list, line_arg='metric', line_vals=line_vals, line_names=line_names,  
        styles=[('red', '-'), ('blue', '-'),  
                ('yellow', '-')], ylabel='ms / TFLOPS / GB/s', plot_name='moe-gemm-benchmark', args={  
                    'dtype': dtype, 'routed_weight': routed_weight, 'use_fp8_w8a8': use_fp8_w8a8, 'use_int8_w8a16':  
                    use_int8_w8a16, 'use_int8_w8a8': use_int8_w8a8, 'fp8_type': fp8_type  
                })  
  
    @triton.testing.perf_report([benchmark])  
    def bench_moe_gemm(M, N, K, E, top_k, dtype, routed_weight, metric, use_fp8_w8a8, use_int8_w8a16, use_int8_w8a8,  
                       fp8_type, model=None):  
        a, b, c, metadata = input_helper(M, N, K, top_k, E, routed_weight=routed_weight, use_fp8_w8a8=use_fp8_w8a8,  
                                         use_int8_w8a16=use_int8_w8a16, use_int8_w8a8=use_int8_w8a8, fp8_type=fp8_type,  
                                         dtype=dtype)  
  
        # (M, K) * (top_k, N, K) -> (M, top_k, N). 2 for multiplication and accumulation  
        flops = 2.0 * M * top_k * K * N  
        # The weight is applied on the gemm product which has the shape of (M, top_k, N)  
        if routed_weight:  
            flops += M * top_k * N  
  
        if use_fp8_w8a8:  
            a_bytes = b_bytes = torch.tensor([], dtype=fp8_type).element_size()  
            c_bytes = torch.tensor([], dtype=dtype).element_size()  
        if use_int8_w8a8: # This should be elif  
            a_bytes = b_bytes = torch.tensor([], dtype=torch.int8).element_size()  
            c_bytes = torch.tensor([], dtype=torch.int8).element_size() # This was torch.int8, should be dtype for c  
        elif use_int8_w8a16:  
            b_bytes = torch.tensor([], dtype=torch.int8).element_size()  
            a_bytes = c_bytes = torch.tensor([], dtype=dtype).element_size()  
        else:  
            a_bytes = b_bytes = c_bytes = torch.tensor([], dtype=dtype).element_size()  
  
        # (M, K) memory load for A (E,  N,  K) for B not (top_k,  N,  K) because we are in total bringing in all expert matrices into the chip from memory. It's just that not all multiply the same A.  
        mem_read = (M * K) * a_bytes + (E * N * K) * b_bytes  
        # Memory write for the gemm product  
        mem_write = (M * top_k * N) * c_bytes  
        mem = mem_read + mem_write  
        fn = lambda: moe_gemm(a, b, c, metadata)  
        ms = triton.testing.do_bench(fn)  
  
        bandwidth = mem / (ms * 1e-3) * 1e-9  # GB/s  
        tflops = flops / ms * 1e-9  

        ############ BENCHMARK LOGGING ############
        log_entry = f"ms: {ms} bandwidth: {bandwidth} tflops: {tflops}" + '\n'
        OUTPUT_FILENAME = __file__.replace('.','_') + '_benchmark.log'
        with open(OUTPUT_FILENAME, 'a', encoding='utf-8') as f:
            f.write(log_entry)
        ############ BENCHMARK LOGGING ############
        
        # Return exactly one scalar depending on which metric is active  
        if metric == 'time':  
            return ms  
        elif metric == 'tflops':  
            return tflops  
        elif metric == 'bandwidth':  
            return bandwidth  
        else:  
            raise ValueError("Unknown metric: " + metric)  
  
    bench_moe_gemm.run(save_path=".", print_data=True)  
  
  
def parse_args():  
    parser = argparse.ArgumentParser(  
        prog="Benchmark MoE GEMM",  
        allow_abbrev=False,  
    )  
    parser.add_argument('-model_configs', type=str, default="model_configs.json", help="Model config json file.")  
    available_models = get_available_models(model_families=["mistral"])  # Dynamically load model names  
    model_help = ("Model name to benchmark. Select from: [" + ", ".join(available_models) +  
                  "]. Use 'all' to benchmark all models or leave blank for the default benchmark script.")  
    parser.add_argument('-model', type=str, default=None, help=model_help)  
    parser.add_argument("-M", type=int, default=0, help="M dimension")  
    parser.add_argument("-K", type=int, default=0, help="K dimension")  
    parser.add_argument("-N", type=int, default=0, help="N dimension")  
    parser.add_argument("-E", type=int, default=0, help="Number of experts")  
    parser.add_argument("-top_k", type=int, default=0, help="top_k experts per token")  
    parser.add_argument("-routed_weight", action='store_true', default=False)  
    parser.add_argument("-int8_w8a16", action='store_true', default=False)  
    parser.add_argument("-int8_w8a8", action='store_true', default=False)  
    parser.add_argument("-fp8_w8a8", action='store_true', default=False)  
    parser.add_argument("-dtype", default='fp16')  
    parser.add_argument("-fp8_type", default='e5m2fnuz')  
    args = parser.parse_args()  
    return args  
  
  
arg_to_torch_dtype = {  
    'fp16': torch.float16, 'bf16': torch.bfloat16, 'fp32': torch.float32, "e5m2fnuz": torch.float8_e5m2fnuz, "e4m3fnuz":  
    torch.float8_e4m3fnuz  
}  
  
  
def main():  
    args = parse_args()  
    custom_config = False  
    # If user provides all M,K,N,E,top_k we consider it custom  
    if args.M and args.K and args.N and args.E and args.top_k:  
        custom_config = True  
    run_benchmark(custom_config, args)

if __name__ == '__main__':  
    sys.exit(main())  



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
from triton import language as tl
import torch


@triton.jit
def mul2_kernel(
    in_ptr0,
    out_ptr,
    n_elements,
    BLOCK_SIZE: "tl.constexpr",
):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(in_ptr0 + offsets, mask=mask)
    output = 2 * x
    tl.store(out_ptr + offsets, output, mask=mask)

@triton.jit
def mul2_inplace_kernel(
    ptr,
    n_elements,
    BLOCK_SIZE: "tl.constexpr",
):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(ptr + offsets, mask=mask)
    output = 2 * x
    tl.store(ptr + offsets, output, mask=mask)


def triton_mul2(x, BLOCK_SIZE=16):
    output = torch.zeros_like(x)
    n_elements = x.numel()
    grid = (triton.cdiv(n_elements, BLOCK_SIZE),)
    mul2_kernel[grid](x, output, n_elements, BLOCK_SIZE)
    return output

def triton_mul2_inplace(x, BLOCK_SIZE=16):
    n_elements = x.numel()
    grid = (triton.cdiv(n_elements, BLOCK_SIZE),)
    mul2_inplace_kernel[grid](x, n_elements, BLOCK_SIZE)
    return x




##################################################################################################################################################


def test_mul():
    N = 1024 * 1024
    x = torch.randn(N, device='cuda')

    # 分支1: triton_mul2 with BLOCK_SIZE=1024
    triton_mul2_result = triton_mul2(x, BLOCK_SIZE=1024)

    # 分支2: triton_mul2_inplace with BLOCK_SIZE=1024
    triton_mul2_inplace_result = triton_mul2_inplace(x.clone(), BLOCK_SIZE=1024)

    # 分支3: triton_mul2 with a different BLOCK_SIZE
    triton_mul2_result_case2 = triton_mul2(x, BLOCK_SIZE=512)

    # 分支4: triton_mul2_inplace with a different BLOCK_SIZE
    triton_mul2_inplace_result_case2 = triton_mul2_inplace(x.clone(), BLOCK_SIZE=512)

    # 返回测试结果
    result_dict = {
        "test_case_1": triton_mul2_result,
        "test_case_2": triton_mul2_inplace_result,
        "test_case_3": triton_mul2_result_case2,
        "test_case_4": triton_mul2_inplace_result_case2,
    }
    
    return result_dict

# 执行测试函数
result_gold = test_mul()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
import triton.language as tl
import torch

@triton.jit
def kernel_f8_to_f16(Y, X, N, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(0)
    offs = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offs < N
    x = tl.load(X + offs, mask=mask)
    tl.store(Y + offs, x, mask=mask)
    tl.store(Y + offs, x, mask=mask)

def f8_to_f16(x, dtypes=tl.float8e5) -> torch.Tensor:
    assert x.dtype == torch.int8, f"torch.int8 expected but got {x.dtype}"
    assert "cuda" in str(x.device), f"CUDA tensors only but got {x.device}"
    ret = torch.empty_like(x, dtype=torch.float16)
    grid = lambda META: (triton.cdiv(x.numel(), META['BLOCK_SIZE']),)
    numel = ret.untyped_storage().size() // ret.element_size()
    kernel_f8_to_f16[grid](ret, triton.reinterpret(x, dtypes), numel, BLOCK_SIZE=1024)
    return ret

@triton.jit
def kernel_f16_to_f8(Y, X, N, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(0)
    offs = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offs < N
    x = tl.load(X + offs, mask=mask)
    tl.store(Y + offs, x, mask=mask)

def f16_to_f8(x: torch.Tensor, dtypes=tl.float8e5) -> torch.Tensor:
    assert x.dtype in [torch.float16, torch.float32]
    assert "cuda" in str(x.device), f"CUDA tensors only but got {x.device}"
    ret = torch.empty_like(x, dtype=torch.int8)
    grid = lambda META: (triton.cdiv(x.numel(), META['BLOCK_SIZE']),)
    numel = x.untyped_storage().size() // x.element_size()
    kernel_f16_to_f8[grid](triton.reinterpret(ret, dtypes), x, numel, BLOCK_SIZE=1024)
    return ret




##################################################################################################################################################


# Test code for f16_to_f8 and f8_to_f16 functions
def test_triton_kernels():
    results = {}
    # Test for f16_to_f8 and f8_to_f16 conversion
    for i in range(4):
        # Create a random tensor of shape (16, 128) with dtype float16 on CUDA
        a = torch.randn((16, 128), dtype=torch.float16, device="cuda")
        
        # Convert from float16 to float8 using f16_to_f8
        b = f16_to_f8(a, dtypes=tl.float8e5)
        
        # Convert back from float8 to float16 using f8_to_f16
        c = f8_to_f16(b, dtypes=tl.float8e5)
        
        # Store results
        results[f'test_case_{i+1}'] = c

    return results

# Run the test and store the result
result_gold = test_triton_kernels()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.dropout_triton import dropout
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('dropout_triton', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 30):
            size = 2 ** i
            x = torch.rand(size, dtype=torch.float16)
            x_keep = torch.rand(size, dtype=torch.float16) > 0.5  # Random mask
            self.input_tensors.append((x, x_keep))

    def to_cuda(self, input_tensor):
        x, x_keep = input_tensor
        return x.cuda(), x_keep.cuda()

    def call_op(self, input_tensor):
        x, x_keep = input_tensor
        p = 0.5  # Example dropout probability
        return dropout(x, x_keep, p)

    def get_gbps(self, input_tensor, runtime):
        x, x_keep = input_tensor
        total_bytes = x.numel() * x.element_size() * 2 + x_keep.numel() * x_keep.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        x, _ = input_tensor
        FLOPS = x.numel()  # Each element involves one division and one conditional operation
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def matmul_kernel(
    x_ptr,
    y_ptr,
    z_ptr,
    m_size,
    k_size,
    n_size,
    m_block_size: tl.constexpr,
    k_block_size: tl.constexpr,
    n_block_size: tl.constexpr,
):
    # Calculate the m and n blocks to compute using the program id
    pid = tl.program_id(0)
    num_n_blocks = tl.cdiv(n_size, n_block_size)
    m_block = pid // num_n_blocks
    n_block = pid % num_n_blocks

    m_offsets = tl.arange(0, m_block_size) + m_block * m_block_size
    n_offsets = tl.arange(0, n_block_size) + n_block * n_block_size
    k_offsets = tl.arange(0, k_block_size)

    x_ptrs = x_ptr + m_offsets[:, None] * k_size + k_offsets[None, :]
    y_ptrs = y_ptr + k_offsets[:, None] * n_size + n_offsets[None, :]
    z_ptrs = z_ptr + m_offsets[:, None] * n_size + n_offsets[None, :]

    z = tl.zeros((m_block_size, n_block_size), dtype=tl.float32)

    for _ in range(0, k_size, k_block_size):
        x_sub = tl.load(x_ptrs)
        y_sub = tl.load(y_ptrs)
        z += tl.dot(x_sub, y_sub, allow_tf32=False)
        x_ptrs += k_block_size
        y_ptrs += k_block_size * n_size

    tl.store(z_ptrs, z)

def matmul(x, y):
    m_size, k_size = x.shape
    _, n_size = y.shape
    z = torch.empty(m_size, n_size, device='cuda:0')
    def grid(meta):
        return (triton.cdiv(m_size, meta['m_block_size']) * triton.cdiv(n_size, meta['n_block_size']),)
    
    matmul_kernel[grid](
        x,
        y,
        z,
        m_size,
        k_size,
        n_size,
        m_size,
        k_size,
        n_size
    )
    return z



##################################################################################################################################################


import torch

def test_matmul():
    # Test the matmul function with different block sizes
    x = torch.randn(16, 16, device='cuda:0')
    y = torch.randn(16, 16, device='cuda:0')

    # Test case 1
    output1 = matmul(x, y)

    return {
        "test_case_1": output1,
    }

result_gold = test_matmul()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

# Backward Triton kernel for Swish-Gated Linear Units (Swiglu)
@triton.autotune(
    configs=[
        triton.Config({'BLOCK_N': 32}),
        triton.Config({'BLOCK_N': 64}),
        triton.Config({'BLOCK_N': 128}),
        triton.Config({'BLOCK_N': 256}),
        triton.Config({'BLOCK_N': 512}),
        triton.Config({'BLOCK_N': 1024}),
    ],
    key=['ncols'],
)
@triton.heuristics({"RECOMPUTE_OUTPUT": lambda args: args["OUT"] is not None})
@triton.jit
def _swiglu_bwd_kernel(
    X, Y, DOUT, OUT, DX, DY, stride_x_row, stride_y_row, stride_dout_row,
    stride_out_row, stride_dx_row, stride_dy_row, ncols, BLOCK_N: tl.constexpr,
    RECOMPUTE_OUTPUT: tl.constexpr
):
    # Map the program id to the row of X and Y it should compute.
    row = tl.program_id(0)
    start_col = tl.program_id(1) * BLOCK_N
    X += row * stride_x_row
    Y += row * stride_y_row
    DOUT += row * stride_dout_row
    if RECOMPUTE_OUTPUT:
        OUT += row * stride_out_row
    DX += row * stride_dx_row
    DY += row * stride_dy_row
    cols = start_col + tl.arange(0, BLOCK_N)
    x = tl.load(X + cols, mask=cols < ncols, other=0.).to(tl.float32)
    y = tl.load(Y + cols, mask=cols < ncols, other=0.).to(tl.float32)
    dout = tl.load(DOUT + cols, mask=cols < ncols, other=0.).to(tl.float32)
    x_sigmoid = tl.sigmoid(x)
    dx = x_sigmoid * (1 + x * (1 - x_sigmoid)) * y * dout
    dy = x * x_sigmoid * dout
    tl.store(DX + cols, dx, mask=cols < ncols)
    tl.store(DY + cols, dy, mask=cols < ncols)
    if RECOMPUTE_OUTPUT:
        out = x * x_sigmoid * y
        tl.store(OUT + cols, out, mask=cols < ncols)

# Function to invoke the backward kernel
def _swiglu_bwd(xy, dout, dxy=None, recompute_output=False, out=None):
    if xy.stride(-1) != 1:
        xy = xy.contiguous()
    if dout.stride(-1) != 1:
        dout = dout.contiguous()
    batch_shape = xy.shape[:-1]
    xy = xy.reshape(-1, xy.shape[-1])
    x, y = xy.chunk(2, dim=-1)
    dout = dout.reshape(-1, dout.shape[-1])
    assert dout.shape == x.shape
    if dxy is None:
        dxy = torch.empty_like(xy)
    else:
        dxy = dxy.reshape(-1, dxy.shape[-1])
        assert dxy.shape == xy.shape
    dx, dy = dxy.chunk(2, dim=-1)
    assert dx.stride(-1) == 1
    assert dy.stride(-1) == 1
    if recompute_output:
        if out is None:
            out = torch.empty_like(x)
        else:
            out = out.reshape(-1, out.shape[-1])
            assert out.shape == x.shape
        assert out.stride(-1) == 1
    M, N = x.shape
    grid = lambda META: (M, triton.cdiv(N, META['BLOCK_N']))
    with torch.cuda.device(x.device.index):
        _swiglu_bwd_kernel[grid](
            x, y, dout, out if recompute_output else None, dx, dy, x.stride(0), y.stride(0),
            dout.stride(0), out.stride(0) if recompute_output else 0, dx.stride(0),
            dy.stride(0), N
        )
    if not recompute_output:
        return dxy.reshape(*batch_shape, dxy.shape[-1])
    else:
        return dxy.reshape(*batch_shape, dxy.shape[-1]), out.reshape(*batch_shape, out.shape[-1])




##################################################################################################################################################


import torch

# Test the backward function
def test_swiglu_bwd():
    # Create random input and gradient tensors
    batch_size = 4
    ncols = 128
    xy = torch.randn(batch_size, 2 * ncols, device='cuda', dtype=torch.float32)
    dout = torch.randn(batch_size, ncols, device='cuda', dtype=torch.float32)
    
    # Call the backward function without recompute_output
    dxy = _swiglu_bwd(xy, dout)
    
    # Call the backward function with recompute_output
    dxy_recompute, out = _swiglu_bwd(xy, dout, recompute_output=True)
    
    # Store results in a dictionary
    results = {
        "test_case_1": dxy,
        "test_case_2": (dxy_recompute, out)
    }
    
    return results

# Run the tests
result_gold = test_swiglu_bwd()



# Copyright(C) [2025] Advanced Micro Devices, Inc. All rights reserved.
import torch
import triton
from typing import Callable, List, Union, Tuple, Dict
import json
import os
import time # For unique filenames if needed

# --- do_bench_config class (can remain largely the same) ---
class do_bench_config():
    def __init__(
        self,
        warm_up: int = 25, # Default values for individual benchmarks
        repetition: int = 100,
        quantiles: List[float] = None,
        return_mode: str = "median"
    ):
        self.warm_up = warm_up
        self.repetition = repetition
        self.quantiles = quantiles if quantiles is not None else [0.5, 0.8, 0.2]
        self.return_mode = return_mode

# --- Global Collector for results ---
# This is a common way to gather data from pytest runs.
# It needs to be managed carefully, especially with parallel pytest (xdist).
# For non-parallel pytest, it's simpler.
# Let's assume non-parallel pytest for now for simplicity of collection.
# If using pytest-xdist, inter-process communication for collection is needed.

PYTEST_BENCHMARK_RESULTS = {} # Key: op_name, Value: list of result dicts

def add_benchmark_result(op_name: str, result_dict: Dict):
    if op_name not in PYTEST_BENCHMARK_RESULTS:
        PYTEST_BENCHMARK_RESULTS[op_name] = []
    PYTEST_BENCHMARK_RESULTS[op_name].append(result_dict)

def save_all_benchmark_results(output_dir: str):
    os.makedirs(output_dir, exist_ok=True)
    for op_name, results_list in PYTEST_BENCHMARK_RESULTS.items():
        if results_list: # Only save if there are results
            file_name = op_name + ".json" # e.g., add_kernel.json
            file_path = os.path.join(output_dir, file_name)
            try:
                with open(file_path, 'w', encoding='utf8') as f:
                    json.dump(results_list, f, indent=4, ensure_ascii=False)
                print(f"Benchmark results for {op_name} saved to: {file_path}")
            except IOError as e:
                print(f"Error saving results for {op_name} to {file_path}: {e}")
    PYTEST_BENCHMARK_RESULTS.clear() # Clear after saving

# --- Simplified Benchmarking Helper ---
class PytestBenchmarker:
    def __init__(self, op_callable: Callable, op_name: str, config: do_bench_config = None):
        self.op_callable = op_callable # This will be the kernel launch lambda
        self.op_name = op_name # e.g., "add_kernel" or "test_add_SIZE_BLOCK_SIZE_dtype"
        self.config = config if config else do_bench_config()

    def run_benchmark(self, current_params_dict: Dict,
                      gbps_calculator: Callable = None, # (inputs, ms) -> gbps
                      tflops_calculator: Callable = None # (inputs, ms) -> tflops
                     ):
        """
        Runs the benchmark for the op_callable (which should be a lambda with inputs already bound).
        current_params_dict: Dictionary describing the current pytest parameters.
        gbps_calculator: A function that takes (original_inputs_tuple, ms) and returns GB/s.
        tflops_calculator: A function that takes (original_inputs_tuple, ms) and returns TFLOPS.
        """
        try:
            ms, min_ms, max_ms = triton.testing.do_bench(
                self.op_callable,
                warmup=self.config.warm_up,
                rep=self.config.repetition,
                quantiles=self.config.quantiles,
                return_mode=self.config.return_mode
            )

            gbps = "N/A"
            if gbps_calculator:
                try:
                    # gbps_calculator needs access to the input tensors that define the problem size
                    # The caller of run_benchmark will need to provide these if they are not part of current_params_dict
                    # For simplicity, let's assume current_params_dict contains enough info, or calculator can access them
                    gbps = gbps_calculator(current_params_dict, ms)
                except Exception as e_gbps:
                    print(f"Warning: GB/s calculation failed for {self.op_name} with params {current_params_dict}: {e_gbps}")


            tflops = "N/A"
            if tflops_calculator:
                try:
                    tflops = tflops_calculator(current_params_dict, ms)
                except Exception as e_tflops:
                    print(f"Warning: TFLOPS calculation failed for {self.op_name} with params {current_params_dict}: {e_tflops}")


            result = {
                "params": current_params_dict, # From pytest.mark.parametrize
                "ms": round(ms, 4),
                "min_ms": round(min_ms, 4),
                "max_ms": round(max_ms, 4),
                "GB/s": round(gbps, 2) if isinstance(gbps, (float, int)) else gbps,
                "TFLOPS": round(tflops, 2) if isinstance(tflops, (float, int)) else tflops,
            }
            # Use a unique op_name for each kernel/function being benchmarked
            add_benchmark_result(self.op_name, result)
            # print(f"Benchmarked {self.op_name} with {current_params_dict}: {ms:.4f} ms")
            return result

        except Exception as e:
            print(f"Error during benchmark for {self.op_name} with params {current_params_dict}: {e}")
            error_result = {
                "params": current_params_dict,
                "error": str(e)
            }
            add_benchmark_result(self.op_name, error_result)
            return error_result


# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.autotune(
    configs=[
        triton.Config({}, num_warps=1),
        triton.Config({}, num_warps=2),
        triton.Config({}, num_warps=4),
        triton.Config({}, num_warps=8),
    ],
    key=["BC", "BK"],
)
@triton.jit
def chunk_gla_fwd_A_kernel_intra_sub_inter(
    q,
    k,
    g,
    A,
    s_k_h,
    s_k_t,
    scale,
    T: tl.constexpr,
    K: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,
    BK: tl.constexpr,
    NC: tl.constexpr
):
    i_t, i_c, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_i, i_j = i_c // NC, i_c % NC
    if i_t * BT + i_i * BC >= T:
        return
    if i_i <= i_j:
        return

    b_A = tl.zeros([BC, BC], dtype=tl.float32)
    for i_k in range(tl.cdiv(K, BK)):
        o_k = i_k * BK + tl.arange(0, BK)
        m_k = o_k < K

        p_q = tl.make_block_ptr(q + i_bh * s_k_h, (T, K), (s_k_t, 1), (i_t * BT + i_i * BC, i_k * BK), (BC, BK), (1, 0))
        p_g = tl.make_block_ptr(g + i_bh * s_k_h, (T, K), (s_k_t, 1), (i_t * BT + i_i * BC, i_k * BK), (BC, BK), (1, 0))
        p_k = tl.make_block_ptr(k + i_bh * s_k_h, (K, T), (1, s_k_t), (i_k * BK, i_t * BT + i_j * BC), (BK, BC), (0, 1))
        p_gk = tl.make_block_ptr(g + i_bh * s_k_h, (K, T), (1, s_k_t), (i_k * BK, i_t * BT + i_j * BC), (BK, BC), (0, 1))
        p_gn = tl.max_contiguous(tl.multiple_of(g + i_bh * s_k_h + (i_t * BT + i_i * BC) * K + o_k, BK), BK)
        b_gn = tl.load(p_gn, mask=m_k, other=0)
        b_q = tl.load(p_q, boundary_check=(0, 1))
        b_g = tl.load(p_g, boundary_check=(0, 1))
        b_qg = b_q * tl.exp(b_g - b_gn[None, :]) * scale
        b_k = tl.load(p_k, boundary_check=(0, 1))
        b_gk = tl.load(p_gk, boundary_check=(0, 1))
        b_kg = b_k * tl.exp(b_gn[:, None] - b_gk)
        b_A += tl.dot(b_qg, b_kg)

    p_A = tl.make_block_ptr(A + i_bh * T * BT, (T, BT), (BT, 1), (i_t * BT + i_i * BC, i_j * BC), (BC, BC), (1, 0))
    tl.store(p_A, b_A.to(A.dtype.element_ty), boundary_check=(0, 1))

@triton.autotune(
    configs=[
        triton.Config({}, num_warps=1),
        triton.Config({}, num_warps=2),
        triton.Config({}, num_warps=4),
        triton.Config({}, num_warps=8),
    ],
    key=["BK", "BT"],
)
@triton.jit
def chunk_gla_fwd_A_kernel_intra_sub_intra(
    q,
    k,
    g,
    A,
    s_k_h,
    s_k_t,
    scale,
    T: tl.constexpr,
    K: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,
    BK: tl.constexpr
):
    i_t, i_i, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_j = i_i
    if i_t * BT + i_i * BC >= T:
        return

    o_i = tl.arange(0, BC)
    o_k = tl.arange(0, BK)
    o_A = i_bh * T * BT + (i_t * BT + i_i * BC + tl.arange(0, BC)) * BT + i_j * BC
    m_k = o_k < K
    m_A = (i_t * BT + i_i * BC + tl.arange(0, BC)) < T

    p_q = tl.make_block_ptr(q + i_bh * s_k_h, (T, K), (s_k_t, 1), (i_t * BT + i_i * BC, 0), (BC, BK), (1, 0))
    p_g = tl.make_block_ptr(g + i_bh * s_k_h, (T, K), (s_k_t, 1), (i_t * BT + i_i * BC, 0), (BC, BK), (1, 0))

    p_k = tl.max_contiguous(tl.multiple_of(k + i_bh * s_k_h + (i_t * BT + i_j * BC) * K + o_k, BK), BK)
    p_gk = tl.max_contiguous(tl.multiple_of(g + i_bh * s_k_h + (i_t * BT + i_j * BC) * K + o_k, BK), BK)

    b_q = tl.load(p_q, boundary_check=(0, 1))
    b_g = tl.load(p_g, boundary_check=(0, 1))
    for j in range(0, min(BC, T-i_t*BT-i_i*BC)):
        b_k = tl.load(p_k, mask=m_k, other=0).to(tl.float32)
        b_gk = tl.load(p_gk, mask=m_k, other=0).to(tl.float32)
        b_A = tl.sum(b_q * b_k[None, :] * tl.exp(b_g - b_gk[None, :]), 1)
        b_A = tl.where(o_i >= j, b_A * scale, 0.)
        tl.store(A + o_A + j, b_A, mask=m_A)
        p_k += K
        p_gk += K

@triton.autotune(
    configs=[
        triton.Config({}, num_warps=1),
        triton.Config({}, num_warps=2),
        triton.Config({}, num_warps=4),
        triton.Config({}, num_warps=8),
    ],
    key=["BC", "BK"],
)
@triton.jit
def chunk_gla_fwd_A_kernel_intra_sub_intra_split(
    q,
    k,
    g,
    A,
    s_k_h,
    s_k_t,
    scale,
    T: tl.constexpr,
    K: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,
    BK: tl.constexpr,
    NC: tl.constexpr
):
    i_k, i_tc, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_t, i_i = i_tc // NC, i_tc % NC
    i_j = i_i
    n_bh = tl.num_programs(2)
    if i_t * BT + i_i * BC >= T:
        return

    o_i = tl.arange(0, BC)
    o_k = i_k * BK + tl.arange(0, BK)
    o_A = (i_bh + i_k * n_bh) * T * BC + (i_t * BT + i_i * BC + tl.arange(0, BC)) * BC
    m_k = o_k < K
    m_A = (i_t * BT + i_i * BC + tl.arange(0, BC)) < T

    p_q = tl.make_block_ptr(q + i_bh * s_k_h, (T, K), (s_k_t, 1), (i_t * BT + i_i * BC, i_k * BK), (BC, BK), (1, 0))
    p_g = tl.make_block_ptr(g + i_bh * s_k_h, (T, K), (s_k_t, 1), (i_t * BT + i_i * BC, i_k * BK), (BC, BK), (1, 0))
    p_k = tl.max_contiguous(tl.multiple_of(k + i_bh * s_k_h + (i_t * BT + i_j * BC) * K + o_k, BK), BK)
    p_gk = tl.max_contiguous(tl.multiple_of(g + i_bh * s_k_h + (i_t * BT + i_j * BC) * K + o_k, BK), BK)

    b_q = tl.load(p_q, boundary_check=(0, 1))
    b_g = tl.load(p_g, boundary_check=(0, 1))
    for j in range(0, min(BC, T-i_t*BT-i_i*BC)):
        b_A = tl.zeros([BC], dtype=tl.float32)
        b_k = tl.load(p_k, mask=m_k, other=0).to(tl.float32)
        b_gk = tl.load(p_gk, mask=m_k, other=0).to(tl.float32)
        b_A += tl.sum(b_q * b_k[None, :] * tl.exp(b_g - b_gk[None, :]), 1)
        b_A = tl.where(o_i >= j, b_A * scale, 0.)
        tl.store(A + o_A + j, b_A, mask=m_A)
        p_k += K
        p_gk += K

@triton.autotune(
    configs=[
        triton.Config({}, num_warps=1),
        triton.Config({}, num_warps=2),
        triton.Config({}, num_warps=4),
        triton.Config({}, num_warps=8),
    ],
    key=["BC"],
)
@triton.jit
def chunk_gla_fwd_A_kernel_intra_sub_intra_merge(
    A,
    A2,
    T: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,
    NK: tl.constexpr
):
    i_t, i_c, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    if i_t * BT + i_c * BC >= T:
        return
    n_bh = tl.num_programs(2)
    b_A = tl.zeros([BC, BC], dtype=tl.float32)
    for i_k in range(0, NK):
        p_A = tl.make_block_ptr(A + (i_bh + i_k*n_bh) * T * BC, (T, BC), (BC, 1), (i_t * BT + i_c * BC, 0), (BC, BC), (1, 0))
        b_A += tl.load(p_A, boundary_check=(0, 1))
    p_A2 = tl.make_block_ptr(A2 + i_bh * T * BT, (T, BT), (BT, 1), (i_t * BT + i_c * BC, i_c * BC), (BC, BC), (1, 0))
    tl.store(p_A2, b_A.to(A2.dtype.element_ty), boundary_check=(0, 1))

@triton.autotune(
    configs=[
        triton.Config({}, num_warps=1),
        triton.Config({}, num_warps=2),
        triton.Config({}, num_warps=4),
        triton.Config({}, num_warps=8),
    ],
    key=["BK", "BV", "BT"],
)
@triton.jit
def chunk_gla_fwd_kernel_o(
    q,
    v,
    g,
    h,
    o,
    A,
    s_k_h,
    s_k_t,
    s_v_h,
    s_v_t,
    s_h_h,
    s_h_t,
    scale,
    T: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr
):
    i_v, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    m_s = tl.arange(0, BT)[:, None] >= tl.arange(0, BT)[None, :]

    b_o = tl.zeros([BT, BV], dtype=tl.float32)
    for i_k in range(tl.cdiv(K, BK)):
        p_q = tl.make_block_ptr(q + i_bh * s_k_h, (T, K), (s_k_t, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0))
        p_g = tl.make_block_ptr(g + i_bh * s_k_h, (T, K), (s_k_t, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0))
        p_h = tl.make_block_ptr(h + i_bh * s_h_h, (K, V), (s_h_t, 1), (i_k * BK, i_v * BV), (BK, BV), (1, 0))

        b_q = tl.load(p_q, boundary_check=(0, 1))
        b_q = (b_q * scale).to(b_q.dtype)
        b_g = tl.load(p_g, boundary_check=(0, 1))
        b_qg = (b_q * tl.exp(b_g)).to(b_q.dtype)
        b_h = tl.load(p_h, boundary_check=(0, 1))
        if i_k >= 0:
            b_o += tl.dot(b_qg, b_h.to(b_qg.dtype))

    p_v = tl.make_block_ptr(v + i_bh * s_v_h, (T, V), (s_v_t, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0))
    p_o = tl.make_block_ptr(o + i_bh * s_v_h, (T, V), (s_v_t, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0))
    p_A = tl.make_block_ptr(A + i_bh * T * BT, (T, BT), (BT, 1), (i_t * BT, 0), (BT, BT), (1, 0))
    b_v = tl.load(p_v, boundary_check=(0, 1))
    b_A = tl.load(p_A, boundary_check=(0, 1))
    b_A = tl.where(m_s, b_A, 0.).to(b_v.dtype)
    b_o += tl.dot(b_A, b_v, allow_tf32=False)
    tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))

def chunk_fwd_intra_gated_gk_fn(q, k, g, scale, BT):
    B, H, T, K = q.shape
    BC = 16
    NC = triton.cdiv(BT, BC)
    NT = triton.cdiv(T, BT)

    BK = min(64, triton.next_power_of_2(K))
    A = q.new_empty(B, H, T, BT, dtype=torch.float32)
    grid = (NT, NC * NC, B * H)
    chunk_gla_fwd_A_kernel_intra_sub_inter[grid](
        q, k, g, A,
        k.stride(1), k.stride(2),
        scale,
        T=T, K=K, BT=BT, BC=BC, BK=BK, NC=NC
    )
    grid = (NT, NC, B * H)
    if K <= 256:
        BK = triton.next_power_of_2(K)
        chunk_gla_fwd_A_kernel_intra_sub_intra[grid](
            q, k, g, A,
            k.stride(1), k.stride(2),
            scale,
            T=T, K=K, BT=BT, BC=BC, BK=BK
        )
    else:
        BK = 128
        NK = triton.cdiv(K, BK)
        A_intra = q.new_empty(NK, B, H, T, BC, dtype=torch.float32)
        grid = (NK, NT * NC, B * H)
        chunk_gla_fwd_A_kernel_intra_sub_intra_split[grid](
            q, k, g, A_intra,
            k.stride(1), k.stride(2),
            scale,
            T=T, K=K, BT=BT, BC=BC, BK=BK, NC=NC
        )
        grid = (NT, NC, B * H)
        chunk_gla_fwd_A_kernel_intra_sub_intra_merge[grid](
            A_intra, A,
            T=T, BT=BT, BC=BC, NK=NK
        )
    return A

def chunk_fwd_o_gated_gk_fn(q, v, g_cumsum, A, h, BT, scale):
    B, H, T, K, V = *q.shape, v.shape[-1]
    BK = min(32, triton.next_power_of_2(K))
    BV = min(32, triton.next_power_of_2(V))
    NV = triton.cdiv(V, BV)
    NT = triton.cdiv(T, BT)

    grid = (NV, NT, B * H)
    o = torch.empty_like(v)
    chunk_gla_fwd_kernel_o[grid](
        q, v, g_cumsum, h, o, A,
        q.stride(1), q.stride(2),
        v.stride(1), v.stride(2),
        h.stride(1), h.stride(2),
        scale,
        T=T, K=K, V=V, BT=BT, BK=BK, BV=BV
    )
    return o




##################################################################################################################################################


def test_chunk_gla_fwd():
    # 测试正常的输入规模
    B = 2  # batch size
    H = 2  # number of heads
    T = 128  # sequence length
    K = 256  # key length
    V = 64  # value length
    BT = 16  # block size for T
    BC = 16  # block size for C (head dimension)
    BK = 64  # block size for K (key length)
    scale = 1.0  # scaling factor

    q = torch.randn(B, H, T, K, dtype=torch.float32, device='cuda')
    k = torch.randn(B, H, T, K, dtype=torch.float32, device='cuda')
    v = torch.randn(B, H, T, V, dtype=torch.float32, device='cuda')
    g = torch.randn(B, H, T, K, dtype=torch.float32, device='cuda')
    h = torch.randn(B, H, K, V, dtype=torch.float32, device='cuda')
    
    A = chunk_fwd_intra_gated_gk_fn(q, k, g, scale, BT)
    o = chunk_fwd_o_gated_gk_fn(q, v, g.cumsum(dim=-1), A, h, BT, scale)

    result = {}
    result['test_case_1'] = o.shape

    # 测试 K > 256 的情况
    B = 2
    H = 2
    T = 128
    K = 512  # 设置 K > 256
    V = 64
    BT = 16
    BC = 16
    BK = 128
    scale = 1.0

    q = torch.randn(B, H, T, K, dtype=torch.float32, device='cuda')
    k = torch.randn(B, H, T, K, dtype=torch.float32, device='cuda')
    v = torch.randn(B, H, T, V, dtype=torch.float32, device='cuda')
    g = torch.randn(B, H, T, K, dtype=torch.float32, device='cuda')
    h = torch.randn(B, H, K, V, dtype=torch.float32, device='cuda')

    A = chunk_fwd_intra_gated_gk_fn(q, k, g, scale, BT)
    o = chunk_fwd_o_gated_gk_fn(q, v, g.cumsum(dim=-1), A, h, BT, scale)

    result['test_case_3'] = o.shape

    return result

result_gold = test_chunk_gla_fwd()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.autotune(
    configs=[
        triton.Config({}, num_warps=1),
        triton.Config({}, num_warps=2),
        triton.Config({}, num_warps=4),
        triton.Config({}, num_warps=8),
        triton.Config({}, num_warps=16),
        # triton.Config({}, num_warps=32),
    ],
    key=["BT", "BK", "BV"], 
)
@triton.jit
def chunk_delta_rule_fwd_kernel_h(
    k,
    v,
    d, 
    v_new,
    h,
    initial_state,
    final_state,
    s_qk_h,
    s_qk_t,
    s_qk_d,
    s_vo_h,
    s_vo_t,
    s_vo_d,
    s_h_h,
    s_h_t,
    H: tl.constexpr,
    T: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    NT: tl.constexpr,
    USE_INITIAL_STATE: tl.constexpr,
    STORE_FINAL_STATE: tl.constexpr
):
    i_k, i_v, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)

    b_h = tl.zeros([BK, BV], dtype=tl.float32)

    if USE_INITIAL_STATE:
        p_h0 = tl.make_block_ptr(initial_state + i_bh * K * V, (K, V), (V, 1), (i_k * BK, i_v * BV), (BK, BV), (1, 0))
        b_h = tl.load(p_h0, boundary_check=(0, 1)).to(tl.float32)

    for i_t in range(NT):
        p_h = tl.make_block_ptr(h + i_bh * s_h_h + i_t * K * V, (K, V), (s_h_t, 1), (i_k * BK, i_v * BV), (BK, BV), (1, 0))
        tl.store(p_h, b_h.to(p_h.dtype.element_ty), boundary_check=(0, 1))
        b_h_cumsum = tl.zeros([BK, BV], dtype=tl.float32)
        for i_c in range(tl.cdiv(BT, BC)):
            p_k = tl.make_block_ptr(k + i_bh * s_qk_h, (K, T), (s_qk_d, s_qk_t), (i_k * BK, i_t * BT + i_c * BC), (BK, BC), (0, 1))
            p_d = tl.make_block_ptr(d + i_bh * s_qk_h, (T, K), (s_qk_t, s_qk_d), (i_t * BT + i_c * BC, i_k * BK), (BC, BK), (1, 0))
            p_v = tl.make_block_ptr(v + i_bh * s_vo_h, (T, V), (s_vo_t, s_vo_d), (i_t * BT + i_c * BC, i_v * BV), (BC, BV), (1, 0))
            p_v_new = tl.make_block_ptr(v_new + i_bh * s_vo_h, (T, V), (s_vo_t, s_vo_d), (i_t * BT + i_c * BC, i_v * BV), (BC, BV), (1, 0))   
            b_k = tl.load(p_k, boundary_check=(0, 1))
            b_d = tl.load(p_d, boundary_check=(0, 1))
            b_v = tl.load(p_v, boundary_check=(0, 1))
            b_v -= tl.dot(b_d, b_h.to(b_k.dtype), allow_tf32=False)
            tl.store(p_v_new, b_v.to(p_v_new.dtype.element_ty), boundary_check=(0, 1))
            b_h_cumsum += tl.dot(b_k, b_v.to(b_k.dtype), allow_tf32=False)
        b_h += b_h_cumsum      
        
    if STORE_FINAL_STATE:
        p_ht = tl.make_block_ptr(final_state + i_bh * K * V, (K, V), (V, 1), (i_k * BK, i_v * BV), (BK, BV), (1, 0))
        tl.store(p_ht, b_h.to(p_ht.dtype.element_ty), boundary_check=(0, 1))

def chunk_fwd_h_fn(k, w, u, BT, initial_state, final_state):
    B, H, T, K, V = *k.shape, u.shape[-1]

    BK = triton.next_power_of_2(K)
    assert BK <= 256, "current kernel does not support head dimension larger than 256."
    BV = 16//2 if BK > 128 else 32//2
    BV = 64//2 if BK <= 64 else BV//2
    BC = 16//2 if BK > 128 else 32 //2
    BC = 64//2 if BK <= 64 else BC//2
    BC = min(BT, BC)
    NT, NK, NV = triton.cdiv(T, BT), triton.cdiv(K, BK), triton.cdiv(V, BV)
    assert NK == 1, 'NK > 1 is not supported because it involves time-consuming synchronization'

    h = k.new_empty(B, H, NT * K, V)
    grid = (NK, NV, B * H)
    v_new = torch.empty_like(u)
    chunk_delta_rule_fwd_kernel_h[grid](
        k, u, w, v_new, h, initial_state, final_state,
        k.stride(1), k.stride(2), k.stride(3),
        u.stride(1), u.stride(2), u.stride(3),
        h.stride(1), h.stride(2),
        H=H, T=T, K=K, V=V, BT=BT, BC=BC, BK=BK, BV=BV, NT=NT,
        USE_INITIAL_STATE=initial_state is not None,
        STORE_FINAL_STATE=final_state is not None,
        )
    return h, v_new




##################################################################################################################################################


import torch

# Test function for chunk_fwd_h_fn
def test_chunk_fwd_h_fn():
    B, H, T, K, V = 2, 4, 128, 64, 64  # Example dimensions
    BT = 32  # Block size for T dimension

    k = torch.randn(B, H, K, T, dtype=torch.float32, device='cuda')
    w = torch.randn(B, H, T, K, dtype=torch.float32, device='cuda')
    u = torch.randn(B, H, T, V, dtype=torch.float32, device='cuda')

    results = {}

    # Test without initial and final states
    h, v_new = chunk_fwd_h_fn(k, w, u, BT, initial_state=None, final_state=None)
    results['test_case_1'] = (h.shape, v_new.shape)

    # Test with initial and final states
    initial_state = torch.zeros(B, H, K, V, dtype=torch.float32, device='cuda')
    final_state = torch.zeros(B, H, K, V, dtype=torch.float32, device='cuda')
    h, v_new = chunk_fwd_h_fn(k, w, u, BT, initial_state=initial_state, final_state=final_state)
    results['test_case_2'] = (h.shape, v_new.shape)

    return results

# Run tests
result_gold = test_chunk_fwd_h_fn()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.fused_rotary_embedding import decoding_fused_rotary_embedding
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('fused_rotary_embedding', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 8):  # Adjust the range as needed for your testing
            total_tokens = 2 ** i
            head_num = 8  # Example head number
            head_dim = 64  # Example head dimension
            kv_head_num = head_num // 2  # Example kv head number
            block_size = 16  # Example block size
            max_position_len = 512  # Example max position length
            bsz = 4  # Example batch size
            max_blocks_per_sequence = 10  # Example max blocks per sequence

            q = torch.rand(total_tokens, head_num, head_dim, dtype=torch.float16)
            k = torch.rand(total_tokens, kv_head_num, head_dim, dtype=torch.float16)
            v = torch.rand(total_tokens, kv_head_num, head_dim, dtype=torch.float16)
            cos = torch.rand(max_position_len, head_dim, dtype=torch.float16)
            sin = torch.rand(max_position_len, head_dim, dtype=torch.float16)
            k_cache = torch.rand(max_blocks_per_sequence, kv_head_num, block_size, head_dim, dtype=torch.float16)
            v_cache = torch.rand(max_blocks_per_sequence, kv_head_num, block_size, head_dim, dtype=torch.float16)
            block_tables = torch.randint(0, max_blocks_per_sequence, (bsz, max_blocks_per_sequence), dtype=torch.int32)
            kv_lengths = torch.randint(1, total_tokens, (bsz,), dtype=torch.int32)

            self.input_tensors.append((q, k, v, cos, sin, k_cache, v_cache, block_tables, kv_lengths))

    def to_cuda(self, input_tensor):
        return tuple(tensor.cuda() for tensor in input_tensor)

    def call_op(self, input_tensor):
        q, k, v, cos, sin, k_cache, v_cache, block_tables, kv_lengths = input_tensor
        return decoding_fused_rotary_embedding(q, k, v, cos, sin, k_cache, v_cache, block_tables, kv_lengths)

    def get_gbps(self, input_tensor, runtime):
        q, k, v, cos, sin, k_cache, v_cache, block_tables, kv_lengths = input_tensor
        total_bytes = (q.numel() + k.numel() + v.numel() + cos.numel() + sin.numel() +
                       k_cache.numel() + v_cache.numel() + block_tables.numel() + kv_lengths.numel()) * q.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        q, k, v, cos, sin, k_cache, v_cache, block_tables, kv_lengths = input_tensor
        # Assuming each element operation is a FLOP
        FLOPS = 2 * (q.numel() + k.numel() + v.numel())
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
import triton.language as tl

inv_ln2 = tl.constexpr(1.44269504)

# Forward decay cumulative sum kernel
@triton.jit
def fwd_decay_cumsum(
    g,
    g_o, 
    s_qk_h,
    s_qk_t,
    s_qk_d,
    B,
    H,
    T,
    scale,
    BT: tl.constexpr,
    BK: tl.constexpr,
    DK: tl.constexpr
):
    i_k, i_c, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    p_g = g + i_bh * s_qk_h + i_c * BT * DK + i_k * BK + tl.arange(0, BK)
    p_go = g_o + i_bh * s_qk_h + i_c * BT * DK + i_k * BK + tl.arange(0, BK)
    cum_decay = tl.zeros([BK], dtype=tl.float32)
    mask = (i_k * BK + tl.arange(0, BK)) < DK

    for i in range(BT):
        _g = tl.load(p_g, mask=mask, other=0).to(tl.float32)
        cum_decay += _g * inv_ln2
        tl.store(p_go, cum_decay.to(p_go.dtype.element_ty), mask=mask)
        p_g += DK
        p_go += DK

# Prepare qg and kg kernel
@triton.jit
def prepare_qg_kg(
    q,
    k,
    g,
    qg,
    kg,
    s_qk_h,
    s_qk_t,
    s_qk_d,
    B,
    H,
    T,
    scale,
    BT: tl.constexpr,
    BK: tl.constexpr,
    DK: tl.constexpr
):

    i_k, i_c, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    p_q = q + i_bh * s_qk_h + i_c * BT * DK + i_k * BK + tl.arange(0, BK)
    p_g = g + i_bh * s_qk_h + i_c * BT * DK + i_k * BK + tl.arange(0, BK)
    p_k = k + i_bh * s_qk_h + i_c * BT * DK + i_k * BK + tl.arange(0, BK)
    p_qg = qg + i_bh * s_qk_h + i_c * BT * DK + i_k * BK + tl.arange(0, BK)
    p_kg = kg + i_bh * s_qk_h + i_c * BT * DK + i_k * BK + tl.arange(0, BK)
    
    mask = (i_k * BK + tl.arange(0, BK)) < DK

    last_decay = tl.load(g + i_bh * s_qk_h + (i_c * BT + BT - 1) * DK + i_k * BK + tl.arange(0, BK))

    for i in range(BT):
        _q = tl.load(p_q, mask=mask, other=0)
        _k = tl.load(p_k, mask=mask, other=0)
        _g = tl.load(p_g, mask=mask, other=0).to(tl.float32)
        _q *= tl.math.exp2(_g) * scale
        _k *= tl.math.exp2(last_decay - _g)
        tl.store(p_kg, _k.to(p_kg.dtype.element_ty), mask=mask)
        tl.store(p_qg, _q.to(p_qg.dtype.element_ty), mask=mask)
        p_q += DK
        p_g += DK
        p_k += DK
        p_kg += DK
        p_qg += DK

# Backward decay global cumulative sum kernel
@triton.jit
def bwd_decay_global_cumsum(
    dq_inner,
    dq_inter,
    dk_inner,
    dk_inter,
    q, k, g, dg,
    s_qk_h,
    s_qk_t,
    s_qk_d,
    B,
    H,
    T,
    scale,
    BT: tl.constexpr,
    BK: tl.constexpr,
    DK: tl.constexpr
):
    i_k, i_c, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    p_q = q + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (i_c * BT + BT - 1) * DK
    p_k = k + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (i_c * BT + BT - 1) * DK
    p_g = g + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (i_c * BT + BT - 1) * DK
    p_dg = dg + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (i_c * BT + BT - 1) * DK
    p_dq_inner = dq_inner + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (i_c * BT + BT - 1) * DK
    p_dk_inner = dk_inner + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (i_c * BT + BT - 1) * DK
    p_dq_inter = dq_inter + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (i_c * BT + BT - 1) * DK
    p_dk_inter = dk_inter + i_bh * s_qk_h + i_k * BK + tl.arange(0, BK) + (i_c * BT + BT - 1) * DK
    cum_grad_dg = tl.zeros([BK], dtype=tl.float32)
    mask = (i_k * BK + tl.arange(0, BK)) < DK
    last_g = tl.zeros([BK], dtype=tl.float32)
    for j in range(BT-1, -1, -1):
        _g = tl.load(p_g, mask=mask, other=0).to(tl.float32)
        if j == (BT-1):
            last_g = _g
        _dq1 = tl.load(p_dq_inner, mask=mask, other=0)
        _dq2 = tl.load(p_dq_inter, mask=mask, other=0)
        _dq2 *= tl.math.exp2(_g)
        _dq = _dq1 + _dq2
        tl.store(p_dq_inter, _dq, mask=mask)
        _dk1 = tl.load(p_dk_inner, mask=mask, other=0)
        _dk2 = tl.load(p_dk_inter, mask=mask, other=0)
        _dk2 *= tl.math.exp2(last_g - _g)
        _dk = _dk1 + _dk2
        tl.store(p_dk_inter, _dk, mask=mask)
        _q = tl.load(p_q, mask=mask, other=0)
        _k = tl.load(p_k, mask=mask, other=0)
        _dg = _dq * _q - _dk * _k
        cum_grad_dg += _dg
        tl.store(p_dg, cum_grad_dg.to(p_dg.dtype.element_ty), mask=mask)
        p_g -= DK
        p_k -= DK
        p_q -= DK
        p_dq_inner -= DK
        p_dk_inner -= DK
        p_dq_inter -= DK
        p_dk_inter -= DK
        p_dg -= DK

# Define the kernel launch functions
def launch_fwd_decay_cumsum(g, g_o, B, H, T, scale, BT, BK, DK):
    # Calculate strides
    s_qk_h = H * T * DK
    s_qk_t = T * DK
    s_qk_d = DK

    # Launch the kernel
    grid = (DK // BK, T // BT, B * H)
    fwd_decay_cumsum[grid](
        g, g_o, s_qk_h, s_qk_t, s_qk_d, B, H, T, scale, BT=BT, BK=BK, DK=DK
    )

def launch_prepare_qg_kg(q, k, g, qg, kg, B, H, T, scale, BT, BK, DK):
    # Calculate strides
    s_qk_h = H * T * DK
    s_qk_t = T * DK
    s_qk_d = DK

    # Launch the kernel
    grid = (DK // BK, T // BT, B * H)
    prepare_qg_kg[grid](
        q, k, g, qg, kg, s_qk_h, s_qk_t, s_qk_d, B, H, T, scale, BT=BT, BK=BK, DK=DK
    )

def launch_bwd_decay_global_cumsum(dq_inner, dq_inter, dk_inner, dk_inter, q, k, g, dg, B, H, T, scale, BT, BK, DK):
    # Calculate strides
    s_qk_h = H * T * DK
    s_qk_t = T * DK
    s_qk_d = DK

    # Launch the kernel
    grid = (DK // BK, T // BT, B * H)
    bwd_decay_global_cumsum[grid](
        dq_inner, dq_inter, dk_inner, dk_inter, q, k, g, dg, s_qk_h, s_qk_t, s_qk_d, B, H, T, scale, BT=BT, BK=BK, DK=DK
    )




##################################################################################################################################################


import torch

# Test the kernels
def test_kernels():
    # Define parameters
    B, H, T, DK = 2, 2, 4, 8
    scale = 1.0
    BT, BK = 2, 4

    # Create input tensors
    g = torch.randn((B, H, T, DK), dtype=torch.float32, device='cuda')
    g_o = torch.zeros_like(g)
    q = torch.randn((B, H, T, DK), dtype=torch.float32, device='cuda')
    k = torch.randn((B, H, T, DK), dtype=torch.float32, device='cuda')
    qg = torch.zeros_like(q)
    kg = torch.zeros_like(k)
    dq_inner = torch.randn_like(q)
    dq_inter = torch.randn_like(q)
    dk_inner = torch.randn_like(k)
    dk_inter = torch.randn_like(k)
    dg = torch.zeros_like(g)

    # Launch kernels
    launch_fwd_decay_cumsum(g, g_o, B, H, T, scale, BT, BK, DK)
    launch_prepare_qg_kg(q, k, g, qg, kg, B, H, T, scale, BT, BK, DK)
    launch_bwd_decay_global_cumsum(dq_inner, dq_inter, dk_inner, dk_inter, q, k, g, dg, B, H, T, scale, BT, BK, DK)

    # Store outputs for verification
    results = {
        "test_case_1": {
            "g_o": g_o.clone(),
            "qg": qg.clone(),
            "kg": kg.clone(),
            "dg": dg.clone()
        }
    }
    return results

# Run the test
result_gold = test_kernels()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.nested_loops_processing import wrapper_nested3
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('nested_loops_processing', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(12, 24):  # Adjust the range as needed for your testing
            size = 2 ** i
            n_rows = size
            n_cols = 4  # Assuming a fixed number of columns for simplicity
            input_tensor = (n_rows, n_cols)
            self.input_tensors.append(input_tensor)

    def to_cuda(self, input_tensor):
        # The input_tensor here is a tuple of (n_rows, n_cols), no need to move to CUDA
        return input_tensor

    def call_op(self, input_tensor):
        n_rows, n_cols = input_tensor
        wrapper_nested3(n_rows, n_cols)

    def get_gbps(self, input_tensor, runtime):
        n_rows, n_cols = input_tensor
        total_bytes = 2 * n_rows * n_cols * 4  # 2 tensors (input and output), each int32 (4 bytes)
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        n_rows, n_cols = input_tensor
        # Assuming each load/store operation is a FLOP, adjust as necessary
        FLOPS = 3 * n_rows * n_cols  # 3 loads/stores per element
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Copyright(C) [2025] Advanced Micro Devices, Inc. All rights reserved.
######################################## Imports ########################################
import pytest
import torch
import triton
import triton.language as tl
import numpy as np
######################################## Imports ########################################

@triton.jit
def sort_kernel(X, Z, N: tl.constexpr, M: tl.constexpr, descending: tl.constexpr):
    offx = tl.arange(0, M)
    offy = tl.arange(0, N) * M
    off2d = offx[None, :] + offy[:, None]
    x = tl.load(X + off2d)
    x = tl.sort(x, descending=descending)
    tl.store(Z + off2d, x)

##################################################################################################################################################  

import numpy as np
import random
import torch 
import os
import pytest
from numpy.random import RandomState
from geak_eval.perf.ROCm.performance_utils_pytest import PytestBenchmarker, do_bench_config, save_all_benchmark_results
from typing import Dict


result_gold = {}

######################################## HELPERS for Eval ######################################## 
def set_seed(seed: int = 42) -> None:
    """
    Set the random seed for reproducibility across multiple libraries and configure PyTorch for deterministic behavior.

    Args:
        seed (int): The seed value to set. Default is 42.
    """
    # Set seed for Python's built-in random module
    random.seed(seed)
    # Set seed for NumPy
    np.random.seed(seed)
    # Set seed for PyTorch on CPU
    torch.manual_seed(seed)
    # Set seed for PyTorch on all GPUs (if available)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    # Ensure deterministic behavior in PyTorch
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # Set environment variable for hash-based operations
    os.environ['PYTHONHASHSEED'] = str(seed)

######################################## HELPERS for Eval ######################################## 




# Common helper code
def np_dtype_from_str(dtype_str):
    if dtype_str == "bfloat16": return np.float32
    return getattr(np, dtype_str)

def gen_numpy_array_for_torch_conversion(shape, dtype_str, low=0, high=100):
    np_dtype = np_dtype_from_str(dtype_str)
    actual_low = int(low)
    actual_high = int(high)

    if dtype_str == 'bfloat16':
        return np.random.uniform(actual_low, actual_high, shape).astype(np.float32)
    if 'float' in dtype_str:
        return np.random.uniform(actual_low, actual_high, shape).astype(np_dtype)
    elif 'int' in dtype_str:
        if actual_high <= actual_low: actual_high = actual_low + 1
        if np.issubdtype(np_dtype, np.integer):
            iinfo = np.iinfo(np_dtype)
            actual_low = max(iinfo.min, actual_low)
            # For randint, high is exclusive. Cap at iinfo.max, so use iinfo.max + 1 for np.random.randint
            # but ensure it does not overflow if iinfo.max is already the true max value.
            effective_high = min(iinfo.max, actual_high)
            if effective_high <= actual_low: effective_high = actual_low +1 
            if effective_high > iinfo.max : effective_high = iinfo.max # Ensure it doesn't exceed max for dtype for randint high bound
            
            # np.random.randint's high parameter is exclusive
            return np.random.randint(actual_low, effective_high + 1 if effective_high < iinfo.max else effective_high, shape, dtype=np_dtype)


    raise ValueError(f"Unsupported dtype_str for gen_numpy_array_for_torch_conversion: {dtype_str}")

def torch_dtype_from_str(dtype_str): # Also needed for torch.from_numpy().to(torch_dtype)
    if dtype_str == "bfloat16": return torch.bfloat16
    if dtype_str == "float16": return torch.float16
    if dtype_str == "float32": return torch.float32
    # No float64 in original parametrize for sort, but good to have
    if dtype_str == "int32": return torch.int32
    raise ValueError(f"Unsupported dtype_str for torch: {dtype_str}")


@pytest.mark.interpreter
@pytest.mark.parametrize("M_cols, N_rows", [[512, 1], [64, 8], [16, 256], [8, 512]])
@pytest.mark.parametrize("descending", [False, True])
@pytest.mark.parametrize("dtype_str", ['int32', 'float16', 'float32', 'bfloat16'])
def test_sort(N_rows, M_cols, descending, dtype_str, request, device='cuda'):
    set_seed()

    x_np = gen_numpy_array_for_torch_conversion((N_rows, M_cols), dtype_str=dtype_str)
    
    torch_dtype = torch_dtype_from_str(dtype_str)
    x = torch.from_numpy(x_np).to(dtype=torch_dtype, device=device) # Ensure correct torch dtype

    y_ref = torch.sort(x, dim=1, descending=descending)[0]
    z_triton = torch.empty_like(x)

    # Grid is (N_rows,) since each program sorts one row
    grid = (N_rows,)
    sort_kernel[grid](x, z_triton, N_rows, M_cols, descending) # num_warps removed, let triton decide or set default

    result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
    
    ################### save tri_out in result_gold ###################
    test_case_name = request.node.name
    sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
    result_gold[sanitized_key_name] = z_triton.clone().detach().cpu()
    ################################################################### 

    assert torch.allclose(y_ref.float(), z_triton.float(), atol=1e-2, rtol=1e-2), \
        f"Sort mismatch for dtype {dtype_str}, shape ({N_rows}, {M_cols}), descending={descending}.\nReference: {y_ref}\nTriton: {z_triton}"


# --- Python wrapper for the kernel for benchmarking ---
def sort_triton_wrapper(x_tensor, z_buffer, N_const, M_const, descending_const, num_warps_launch):
    # For benchmarking, we assume the kernel processes one N_const x M_const tile.
    # The grid will be (1,) as the kernel is not tiled with program_id for larger inputs.
    grid = (1,) 
    sort_kernel[grid](
        x_tensor, z_buffer, 
        N=N_const, M=M_const, # Pass as constexpr
        descending=descending_const,
        num_warps=num_warps_launch # Launch hint
    )
    return z_buffer

# --- Define TFLOPS and GB/s calculators ---
def calculate_sort_tflops(params: dict, ms: float) -> float:
    # Sorting N_rows * (M_cols * log(M_cols)) comparisons approx.
    # Each comparison is 1 FLOP. This is a rough lower bound.
    N_rows = params['N_tile_rows'] 
    M_cols = params['M_tile_cols']
    if M_cols <= 1: return 0.0 # No comparisons if 0 or 1 element per row
    
    # For tl.sort on each row of M_cols elements
    flops_per_row = M_cols * np.log2(M_cols) if M_cols > 0 else 0 # O(M log M) comparisons
    total_flops = N_rows * flops_per_row
    tflops = total_flops / (ms / 1000) / 1e12
    return tflops

def calculate_sort_gbps(params: dict, ms: float) -> float:
    N_rows = params['N_tile_rows'] 
    M_cols = params['M_tile_cols']
    dtype_str = params.get('dtype_str', 'fp32') 

    current_torch_dtype = torch_dtype_from_str(dtype_str)
    element_size = torch.tensor([], dtype=current_torch_dtype).element_size()

    elements_processed = N_rows * M_cols
    bytes_read = elements_processed * element_size
    bytes_write = elements_processed * element_size 
    total_bytes = bytes_read + bytes_write
    gbps = total_bytes / (ms / 1000) / 1e9
    return gbps

OP_NAME_FOR_BENCHMARK = "triton_sort_perf"

# --- Pytest parametrize for performance testing ---
# Kernel's N and M are constexpr, defining the tile size it processes.
SORT_TILE_SHAPES_FOR_PERF = [
    # N_tile_rows, M_tile_cols (for the single tile processed by kernel)
    (1, 1024), (1, 4096), (1, 8192), (1, 16384), # Sorting long single rows
    (8, 512), (8, 1024), (8, 2048),             # Sorting a few medium rows
    (64, 128), (64, 256),            # Sorting more, shorter rows
    (256, 64), (256, 128), (256, 256)
]
SORT_DTYPES_FOR_PERF = ['int32', 'float16', 'float32', 'bfloat16'] 
SORT_DESCENDING_FOR_PERF = [False, True]
# NUM_WARPS_FOR_PERF = [1, 2, 4, 8]

@pytest.mark.parametrize("N_tile_rows, M_tile_cols", SORT_TILE_SHAPES_FOR_PERF)
@pytest.mark.parametrize("descending_val", SORT_DESCENDING_FOR_PERF)
@pytest.mark.parametrize("dtype_str", SORT_DTYPES_FOR_PERF)
# @pytest.mark.parametrize("num_warps_launch", NUM_WARPS_FOR_PERF) # Optional
def test_performance(N_tile_rows, M_tile_cols, descending_val, dtype_str, request, device='cuda'):
    # num_warps_launch = 4 # Or from parametrize
    
    if dtype_str == 'bfloat16':
        cap = torch.cuda.get_device_capability()
        if cap[0] < 8:
            pytest.skip("bfloat16 requires Ampere+ (arch 80+)")
    
    set_seed()
    current_torch_dtype = torch_dtype_from_str(dtype_str)

    # Input tensor `x_perf` has shape (N_tile_rows, M_tile_cols)
    # as the kernel is designed to process one such tile.
    x_np_perf = gen_numpy_array_for_torch_conversion((N_tile_rows, M_tile_cols), dtype_str)
    x_perf_tensor = torch.from_numpy(x_np_perf).to(dtype=current_torch_dtype, device=device)
    
    z_perf_buffer = torch.empty_like(x_perf_tensor)
    
    op_lambda = lambda: sort_triton_wrapper(
        x_perf_tensor, z_perf_buffer, 
        N_tile_rows, M_tile_cols, descending_val,
        num_warps_launch=4 # Example, can be parametrized
    )

    bench_config = do_bench_config(warm_up=50, repetition=200) # Sorting can vary
    benchmarker = PytestBenchmarker(op_callable=op_lambda,
                                    op_name=OP_NAME_FOR_BENCHMARK,
                                    config=bench_config)
    current_params_for_logs_and_calc = {
        "N_tile_rows": N_tile_rows, 
        "M_tile_cols": M_tile_cols,
        "descending": descending_val,
        "dtype_str": dtype_str,
        "num_warps": 4 # Log the fixed num_warps
    }
    
    perf_result = benchmarker.run_benchmark(current_params_dict=current_params_for_logs_and_calc,
                                            gbps_calculator=calculate_sort_gbps,
                                            tflops_calculator=calculate_sort_tflops)



######################################## HELPERS for Eval ########################################     
# --- Pytest hook to save the dictionary at the end of the session ---  
def test_save_results():  
    """  
    Called after whole test run finished, right before returning the exit status to the system.  
    """
    print('Inside session finish...')
    if "_CALL_SUCCESS_" not in result_gold:
        result_gold['_CALL_SUCCESS_'] = torch.tensor([[0.0]])
    OUTPUT_FILENAME = __file__.replace('.','_') + '.pt'
    print(f"\nSaving all y_triton results to {OUTPUT_FILENAME}...")  
    # Ensure the directory for the output file exists if it's in a subdirectory  
    output_dir = os.path.dirname(OUTPUT_FILENAME)  
    if output_dir and not os.path.exists(output_dir):  
        os.makedirs(output_dir, exist_ok=True)  
    torch.save(result_gold, OUTPUT_FILENAME)       
    print(f"Successfully saved {len(result_gold)} y_triton tensors to {OUTPUT_FILENAME}.")  

def test_save_performance_results():
    """
    Called after the test_performance function finishes.
    This is a separate hook to ensure performance results are saved.
    """
    print('\nPytest session finishing... Saving benchmark results...')

    output_directory = os.path.join(os.path.dirname(__file__), "perf")  # Save in a "perf" subdirectory next to the test file
    os.makedirs(output_directory, exist_ok=True)
    
    save_all_benchmark_results(output_directory)
    print(f"All benchmark results attempted to save to: {output_directory}")


######################################## HELPERS for Eval ########################################


# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
from typing import Dict, Sequence, Union
import torch
import triton
import triton.language as tl


KERNEL_META = dict()

def get_kernel_meta(tensor: torch.Tensor):
    """kernel meta."""
    return KERNEL_META

TypeHintType = Union[Dict[str, type], Sequence[type], None]


@triton.jit
def _multinomial_sampling_kernel(Scores, Seeds, Offsets, Indices, Outputs,
                                 stride_sb, stride_st, stride_ib, stride_it,
                                 num_batchs, num_tokens, BLOCK: tl.constexpr,
                                 BLOCK_N: tl.constexpr):
    """Kernel."""
    batch_block_id = tl.program_id(0)

    off = batch_block_id * BLOCK + tl.arange(0, BLOCK)
    n_off = tl.arange(0, BLOCK_N)

    off_mask = off < num_batchs
    seed = tl.load(Seeds + off, mask=off_mask)
    offset = tl.load(Offsets + off, mask=off_mask).to(tl.int32)

    samp = tl.rand(seed, offset)[:, None]
    acc = tl.zeros((BLOCK, ), dtype=tl.float32)
    output = tl.load(Indices + off * stride_ib, mask=off_mask)

    for b_idx in range(0, num_tokens, BLOCK_N):
        s_off = b_idx + n_off
        s_mask = off_mask[:, None] & (s_off[None, :] < num_tokens)
        scores = tl.load(Scores + off[:, None] * stride_sb +
                         s_off[None, :] * stride_st,
                         mask=s_mask,
                         other=0.0).to(tl.float32)
        c_scores = tl.cumsum(scores, 1)
        cum_scores = acc[:, None] + c_scores
        acc += tl.max(c_scores, 1)

        pre_cum_scores = cum_scores - scores
        valid_mask = (samp > pre_cum_scores) & (samp <= cum_scores)
        found_mask = tl.sum(valid_mask, 1) > 0

        valid_pos = b_idx + tl.argmax(valid_mask.to(tl.int32), 1)
        indices = tl.load(Indices + off * stride_ib + valid_pos * stride_it,
                          mask=found_mask & off_mask,
                          other=-1)
        output = tl.where(found_mask, indices, output)

    tl.store(Outputs + off, output, mask=off_mask)


def multinomial_sampling(scores: torch.Tensor,
                         seeds: torch.LongTensor,
                         offsets: torch.LongTensor,
                         indices: torch.Tensor = None):
    """multinomial sampling."""

    assert scores.dim() == 2
    batch_size, num_tokens = scores.size()
    device = scores.device

    if num_tokens == 1:
        return torch.zeros_like(scores, dtype=torch.long)

    if indices is None:
        indices = torch.arange(num_tokens, device=device)
        indices = indices.expand_as(scores)

    assert indices.dim() == 2
    assert indices.size() == scores.size()

    outputs = indices[:, 0].clone()

    BLOCK = 8
    BLOCK_N = 128

    grid = [triton.cdiv(batch_size, BLOCK)]
    kernel_meta = get_kernel_meta(scores)
    _multinomial_sampling_kernel[grid](scores,
                                       seeds,
                                       offsets,
                                       indices,
                                       outputs,
                                       stride_sb=scores.stride(0),
                                       stride_st=scores.stride(1),
                                       stride_ib=indices.stride(0),
                                       stride_it=indices.stride(1),
                                       num_batchs=batch_size,
                                       num_tokens=num_tokens,
                                       BLOCK=BLOCK,
                                       BLOCK_N=BLOCK_N,
                                       num_warps=8,
                                       **kernel_meta)

    return outputs




##################################################################################################################################################


import torch

def test_multinomial_sampling():
    result_dict = {}

    # Test case 1: Basic functionality with default indices
    scores = torch.tensor([[0.1, 0.2, 0.7], [0.3, 0.4, 0.3]], dtype=torch.float32).cuda()
    seeds = torch.tensor([123, 456], dtype=torch.int64).cuda()
    offsets = torch.tensor([0, 0], dtype=torch.int64).cuda()

    outputs = multinomial_sampling(scores, seeds, offsets)
    result_dict['test_case_1'] = outputs

    # Test case 2: Providing custom indices
    indices = torch.tensor([[0, 1, 2], [2, 1, 0]], dtype=torch.int64).cuda()
    outputs = multinomial_sampling(scores, seeds, offsets, indices)
    result_dict['test_case_2'] = outputs

    # Test case 3: Single token case
    scores_single_token = torch.tensor([[1.0], [1.0]], dtype=torch.float32).cuda()
    outputs = multinomial_sampling(scores_single_token, seeds, offsets)
    result_dict['test_case_3'] = outputs

    return result_dict

result_gold = test_multinomial_sampling()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.matrix_reduction import load_reduce
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('matrix_reduction', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 10):  # Adjust the range as needed for testing
            BLOCK_M = 2 ** i
            BLOCK_N = 2 ** i
            # x = torch.randn((BLOCK_M, 128), dtype=torch.float16)
            self.input_tensors.append((BLOCK_M, BLOCK_N))

    def to_cuda(self, input_tensor):
        return input_tensor

    def call_op(self, input_tensor):
        BLOCK_M, BLOCK_N = input_tensor
        return load_reduce(BLOCK_M, BLOCK_N, 'float16')

    def get_gbps(self, input_tensor, runtime):
        BLOCK_M, BLOCK_N, = input_tensor
        dtype = torch.float16
        x = torch.randn((BLOCK_M, BLOCK_N), dtype=dtype)
        y = torch.empty((BLOCK_M,), dtype=dtype)
        total_bytes = x.numel() * x.element_size() + y.numel() * y.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        BLOCK_M, BLOCK_N = input_tensor
        dtype = torch.float16
        x = torch.randn((BLOCK_M, BLOCK_N), dtype=dtype)
        y = torch.empty((BLOCK_M,), dtype=dtype)
        FLOPS = BLOCK_M * BLOCK_N  # Assuming one operation per element
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()
